import shutil
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple
import tomlkit
import tomlkit.exceptions
import yaml

from dlt.common.runtime import run_context
from dlt._workspace.cli import echo as fmt, utils
from dlt._workspace.cli.utils import DEFAULT_MCP_FEATURES
from dlt._workspace.cli.formatters import parse_frontmatter
from dlt._workspace.cli.exceptions import (
    CliCommandException,
)
from dlt._workspace.cli.ai.agents import AI_AGENTS, TComponentType, _AIAgent, InstallAction
from dlt._workspace.typing import TAiStatusInfo, TToolkitInfo
from dlt._workspace.cli.ai.utils import (
    build_toolkits_dependency_map,
    compute_file_hash,
    extract_toolkit_info,
    fetch_ai_status,
    fetch_secrets_list,
    fetch_secrets_update_fragment,
    fetch_secrets_view_redacted,
    fetch_workbench_toolkit_info,
    fetch_workbench_base,
    is_toolkit_installed,
    load_toolkits_index,
    read_workbench_toolkit_mcp_servers,
    read_workbench_toolkit_combined_info,
    resolve_toolkit_dependencies,
    safe_write_text,
    save_toolkit_entry,
    fetch_workbench_toolkits,
    _INIT_TOOLKIT,
)


@utils.track_command("ai", False, operation="secrets_list")
def ai_secrets_list_command() -> None:
    """Lists project-scoped secret file locations from TOML providers."""
    locations = fetch_secrets_list()
    fmt.echo("Secret file locations:")
    for loc in locations:
        if profile_name := loc.get("profile_name"):
            fmt.echo("  %s (profile: %s)" % (loc["path"], profile_name))
        else:
            fmt.echo("  %s" % loc["path"])


@utils.track_command("ai", False, operation="secrets_view_redacted")
def ai_secrets_view_redacted_command(path: Optional[str] = None) -> None:
    """Prints a redacted secrets TOML."""
    result = fetch_secrets_view_redacted(path)
    if result is None:
        if path:
            fmt.warning("Secrets file not found: %s" % path)
        else:
            fmt.warning("No secrets found in project providers.")
        return
    fmt.echo(result)


@utils.track_command("ai", False, operation="secrets_update_fragment")
def ai_secrets_update_fragment_command(fragment: str, path: str) -> None:
    """Merges a TOML fragment into secrets file and prints the redacted result."""
    try:
        result = fetch_secrets_update_fragment(fragment, path)
    except tomlkit.exceptions.TOMLKitError as ex:
        fmt.error("Invalid TOML fragment: %s" % str(ex))
        raise CliCommandException()
    fmt.echo(result)


@utils.track_command("ai", track_before=True, operation="mcp")
def ai_mcp_run_command(
    port: int = 8000,
    stdio: bool = False,
    sse: bool = False,
    features: Optional[List[str]] = None,
) -> None:
    """Start the dlt MCP server for pipeline data inspection."""
    from dlt._workspace.mcp import WorkspaceMCP

    if stdio:
        transport = "stdio"
    elif sse:
        transport = "sse"
    else:
        transport = "streamable-http"
    from dlt._workspace.mcp.server import resolve_features

    resolved = resolve_features(features)
    fmt.echo(
        "Starting dlt MCP server with features: %s" % ", ".join(sorted(resolved)),
        err=True,
    )
    mcp_server = WorkspaceMCP("dlt", port=port, features=resolved)
    mcp_server.run(transport=transport)


@utils.track_command("ai", track_before=True, operation="mcp-install")
def ai_mcp_install_command(
    agent: Optional[str] = None,
    features: Optional[List[str]] = None,
    name: str = "dlt-workspace",
    overwrite: bool = False,
) -> None:
    """Install dlt MCP server config into the agent's config file."""
    from dlt._workspace.mcp.server import resolve_features

    project_root = Path(run_context.active().run_dir)
    variant = _resolve_agent(agent, project_root)

    resolved = resolve_features(features)
    args = ["uv", "run", "dlt", "ai", "mcp", "run", "--stdio"]
    if resolved != DEFAULT_MCP_FEATURES:
        args.extend(["--features"] + sorted(resolved))

    server_config: Dict[str, Any] = {"command": args[0], "args": args[1:], "type": "stdio"}
    new_servers = {name: server_config}

    config_path = variant.mcp_config_path(project_root)
    existing_content = ""
    if config_path.is_file():
        existing_content = config_path.read_text(encoding="utf-8")

    if not overwrite:
        existing_servers = variant.parse_mcp_servers(existing_content)
        if name in existing_servers:
            fmt.echo("MCP server %s already configured in %s" % (fmt.bold(name), config_path))
            return

    merged = variant.merge_mcp_servers(existing_content, new_servers)
    config_path.parent.mkdir(parents=True, exist_ok=True)
    config_path.write_text(merged, encoding="utf-8")
    fmt.echo("Installed MCP server %s in %s" % (fmt.bold(name), config_path))


def _fetch_workbench_base_cli(location: str, branch: Optional[str]) -> Optional[Path]:
    """Fetch AI workbench repo, warn on failure. CLI wrapper around fetch_workbench_base."""
    try:
        return fetch_workbench_base(location, branch)
    except FileNotFoundError as ex:
        fmt.warning(str(ex))
        return None


def _validate_md_frontmatter(md_file: Path) -> Optional[str]:
    """Validate YAML frontmatter in a markdown file. Returns error message or None."""
    try:
        parse_frontmatter(md_file.read_text(encoding="utf-8"))
    except yaml.YAMLError as ex:
        return "%s: invalid YAML frontmatter: %s" % (md_file.name, ex)
    return None


def _validate_skill_dir(skill_path: Path) -> List[str]:
    """Validate all markdown files in a skill directory. Returns list of errors."""
    errors: List[str] = []
    for md_file in sorted(skill_path.rglob("*.md")):
        err = _validate_md_frontmatter(md_file)
        if err:
            errors.append(err)
    return errors


def _plan_toolkit_install(
    toolkit_dir: Path,
    agent: "_AIAgent",
    project_root: Path,
    toolkit_name: str,
    overwrite: bool = False,
) -> Tuple[List["InstallAction"], List[str]]:
    """Scan toolkit directory and build install actions. Reads source files but does not
    write to project_root. Returns (actions, validation_warnings)."""
    actions: List[InstallAction] = []
    warnings: List[str] = []

    # skills (directory-based, each with SKILL.md + optional files)
    skills_dir = toolkit_dir / "skills"
    if skills_dir.is_dir():
        for skill_path in sorted(skills_dir.iterdir()):
            if not skill_path.is_dir() or not (skill_path / "SKILL.md").exists():
                continue
            errors = _validate_skill_dir(skill_path)
            if errors:
                for err in errors:
                    warnings.append("Skipping skill %s: %s" % (skill_path.name, err))
                continue
            actions.extend(
                agent.install_actions(
                    "skill", skill_path, skill_path.name, toolkit_name, project_root, overwrite
                )
            )

    # ignore file (.claudeignore → agent-specific name)
    ignore_file = toolkit_dir / ".claudeignore"
    if ignore_file.is_file():
        raw_content = ignore_file.read_text(encoding="utf-8")
        actions.extend(
            agent.install_actions(
                "ignore", raw_content, ".claudeignore", toolkit_name, project_root, overwrite
            )
        )

    # commands and rules (markdown files with frontmatter)
    component_types: List[Tuple[TComponentType, str]] = [("command", "commands"), ("rule", "rules")]
    for component_type, dir_name in component_types:
        src_dir = toolkit_dir / dir_name
        if not src_dir.is_dir():
            continue
        for md_file in sorted(src_dir.glob("*.md")):
            err = _validate_md_frontmatter(md_file)
            if err:
                warnings.append("Skipping %s %s: %s" % (component_type, md_file.stem, err))
                continue
            source_name = md_file.stem
            raw_content = md_file.read_text(encoding="utf-8")
            actions.extend(
                agent.install_actions(
                    component_type, raw_content, source_name, toolkit_name, project_root, overwrite
                )
            )

    # mcp server definitions (merge into agent config)
    mcp_servers = read_workbench_toolkit_mcp_servers(toolkit_dir)
    if mcp_servers:
        config_path = agent.mcp_config_path(project_root)
        existing_content = ""
        existing_servers: Dict[str, Any] = {}
        if config_path.is_file():
            existing_content = config_path.read_text(encoding="utf-8")
            existing_servers = agent.parse_mcp_servers(existing_content)

        new_servers = (
            dict(mcp_servers)
            if overwrite
            else {
                name: config for name, config in mcp_servers.items() if name not in existing_servers
            }
        )
        if new_servers:
            merged = agent.merge_mcp_servers(existing_content, new_servers)
            actions.append(
                InstallAction(
                    kind="mcp",
                    source_name=", ".join(sorted(new_servers)),
                    dest_path=config_path,
                    op="save",
                    content_or_path=merged,
                    conflict=False,
                )
            )

    actions = agent.finalize_actions(actions, project_root, workbench_base=toolkit_dir.parent)
    return actions, warnings


def _execute_install(
    actions: List["InstallAction"],
    overwrite: bool = False,
    toolkit_meta: Optional[TToolkitInfo] = None,
    agent_name: Optional[str] = None,
    project_root: Optional[Path] = None,
) -> int:
    """Write non-conflicting actions to disk. Returns count of items installed."""
    installed = 0
    written_paths: List[Tuple[Path, str]] = []  # (dest_path, kind)
    mcp_server_names: List[str] = []
    for action in actions:
        if action.conflict:
            continue
        action.dest_path.parent.mkdir(parents=True, exist_ok=True)
        if action.op == "copytree":
            shutil.copytree(
                action.content_or_path,
                action.dest_path,
                dirs_exist_ok=overwrite,
            )
        else:
            safe_write_text(action.dest_path, action.content_or_path)  # type: ignore[arg-type]
        if action.kind == "mcp":
            # source_name is ", ".join(sorted(new_servers))
            mcp_server_names.extend(s.strip() for s in action.source_name.split(",") if s.strip())
        elif action.skip_index:
            # shared merge targets (e.g. AGENTS.md) — don't track per-toolkit
            pass
        else:
            written_paths.append((action.dest_path, action.op))
        installed += 1

    if installed > 0 and toolkit_meta is not None:
        tracked_files: Optional[Dict[str, Any]] = None
        if project_root is not None:
            tracked_files = {}
            for dest_path, op in written_paths:
                if op == "copytree":
                    for f in sorted(dest_path.rglob("*")):
                        if f.is_file():
                            rel = str(f.relative_to(project_root))
                            tracked_files[rel] = {"sha3_256": compute_file_hash(f)}
                else:
                    rel = str(dest_path.relative_to(project_root))
                    tracked_files[rel] = {"sha3_256": compute_file_hash(dest_path)}
        save_toolkit_entry(
            toolkit_meta,
            agent=agent_name,
            files=tracked_files,
            mcp_servers=sorted(mcp_server_names) if mcp_server_names else None,
        )
    return installed


def _resolve_agent(agent: Optional[str], project_root: Path) -> "_AIAgent":
    """Resolve an explicit agent name or auto-detect. Raises on failure."""
    if agent is not None:
        if agent not in AI_AGENTS:
            fmt.error("Unknown agent: %s" % agent)
            raise CliCommandException()
        return AI_AGENTS[agent]()

    # check init toolkit in .toolkits index for a previously recorded agent
    index = load_toolkits_index()
    init_entry = index.get(_INIT_TOOLKIT)
    if isinstance(init_entry, dict):
        recorded = init_entry.get("agent")
        if recorded and recorded in AI_AGENTS:
            return AI_AGENTS[recorded]()

    detected = _AIAgent.detect_all(project_root)
    if not detected:
        available = ", ".join(sorted(AI_AGENTS))
        fmt.error("Could not detect AI coding agent. Use --agent to specify one of: %s" % available)
        raise CliCommandException()
    best_agent, best_level = detected[0]
    at_best = [a for a, lvl in detected if lvl == best_level]
    if len(at_best) > 1:
        names = ", ".join(a.name for a in at_best)
        fmt.error("Multiple AI coding agents detected: %s. Use --agent to specify one." % names)
        raise CliCommandException()
    fmt.echo("Detected AI coding agent: %s" % fmt.bold(best_agent.name))
    return best_agent


def _report_and_execute(
    actions: List["InstallAction"],
    validation_warnings: List[str],
    overwrite: bool = False,
    strict: bool = False,
    toolkit_meta: Optional[TToolkitInfo] = None,
    agent_name: Optional[str] = None,
    project_root: Optional[Path] = None,
) -> int:
    """Print planned actions, skip conflicts, execute, return count installed.

    In strict mode, validation warnings are treated as errors and raise CliCommandException.
    """
    for w in validation_warnings:
        fmt.warning(w)
    if strict and validation_warnings:
        fmt.error(
            "%d validation error(s). Fix the issues above or install without --strict."
            % len(validation_warnings)
        )
        raise CliCommandException()
    for a in actions:
        if a.conflict:
            fmt.warning(
                "  Skipping %s %s (already exists at %s)" % (a.kind, a.source_name, a.dest_path)
            )
        else:
            fmt.echo("  + %s %s -> %s" % (a.kind, fmt.bold(a.source_name), a.dest_path))
    installed = _execute_install(
        actions,
        overwrite=overwrite,
        toolkit_meta=toolkit_meta,
        agent_name=agent_name,
        project_root=project_root,
    )
    fmt.echo(
        "%s item(s) installed. Please restart your %s session for changes to take effect"
        % (fmt.bold(str(installed)), fmt.bold(str(agent_name)))
    )
    workflow_entry_skill = toolkit_meta.get("workflow_entry_skill") if toolkit_meta else None
    if installed > 0 and workflow_entry_skill:
        fmt.echo("Use %s skill to start!" % fmt.bold(workflow_entry_skill))
    return installed


def _install_toolkit(
    name: str,
    base: Path,
    agent: "_AIAgent",
    project_root: Path,
    overwrite: bool = False,
    strict: bool = False,
) -> None:
    """Core install logic: read plugin.json, version-check, plan, execute."""
    toolkit_dir = base / name
    meta = read_workbench_toolkit_combined_info(toolkit_dir)
    if meta is None:
        fmt.warning(
            "Toolkit %s not found (missing %s)" % (fmt.bold(name), ".claude-plugin/plugin.json")
        )
        return

    toolkit_meta = extract_toolkit_info(meta, name)
    toolkit_name = toolkit_meta["name"]
    toolkit_version = toolkit_meta["version"]

    installed_index = load_toolkits_index()
    local = installed_index.get(toolkit_name)
    if local and not overwrite:
        local_version = local.get("version", "?")
        if local_version == toolkit_version:
            fmt.echo("Toolkit %s %s is already installed." % (toolkit_name, local_version))
        else:
            fmt.echo(
                "Toolkit %s %s is installed, version %s available. Use --overwrite to update."
                % (toolkit_name, local_version, toolkit_version)
            )
        if workflow_entry_skill := toolkit_meta.get("workflow_entry_skill"):
            fmt.echo("Use %s skill to start!" % fmt.bold(workflow_entry_skill))
        return

    actions, warnings = _plan_toolkit_install(
        toolkit_dir, agent, project_root, toolkit_name, overwrite=overwrite
    )
    if not actions and not warnings:
        fmt.echo("No components found in toolkit %s." % fmt.bold(name))
        return

    _report_and_execute(
        actions,
        warnings,
        overwrite=overwrite,
        strict=strict,
        toolkit_meta=toolkit_meta,
        agent_name=agent.name,
        project_root=project_root,
    )


def _install_dependencies(
    name: str,
    toolkits: Dict[str, TToolkitInfo],
    base: Path,
    agent: "_AIAgent",
    project_root: Path,
) -> None:
    """Install upstream dependencies for `name` that are not yet installed."""
    dep_map = build_toolkits_dependency_map(toolkits)
    try:
        deps = resolve_toolkit_dependencies(name, dep_map)
    except ValueError as ex:
        fmt.warning(str(ex))
        return
    for dep in deps:
        if is_toolkit_installed(dep):
            continue
        _install_toolkit(dep, base, agent, project_root)


_WARNING_MESSAGES = {
    "not_initialized": "Workspace not yet initialized (dlt init not yet run)",
    "no_init_toolkit": "MCP server and workflow rules not available (dlt ai init not yet run)",
    "no_toolkits": "No toolkit with workflow is installed!",
    "mcp_unavailable": "MCP server cannot be started due to:",
}


def _print_ai_status(status: TAiStatusInfo) -> None:
    """Render AI status info to the CLI."""
    fmt.echo("dlt %s" % fmt.bold(status["dlt_version"]))
    if status["agent_name"]:
        fmt.echo("Agent: %s" % fmt.bold(status["agent_name"]))

    for code in status["warnings"]:
        msg = _WARNING_MESSAGES.get(code, code)
        fmt.warning(msg)
        if code == "mcp_unavailable" and "mcp_error" in status:
            fmt.echo("  %s" % status["mcp_error"])

    toolkits = status["toolkits"]
    if toolkits:
        fmt.echo("\nInstalled toolkits:")
        for name, entry in sorted(toolkits.items()):
            skill = entry.get("workflow_entry_skill", "")
            if skill:
                fmt.echo("  %s — start with %s skill" % (fmt.bold(name), fmt.bold(skill)))
            else:
                fmt.echo("  %s" % fmt.bold(name))


@utils.track_command("ai", False, operation="status")
def ai_status_command() -> None:
    """Show current AI setup status: dlt version, agent, toolkits, and readiness checks."""
    project_root = Path(run_context.active().run_dir)
    _print_ai_status(fetch_ai_status(project_root))


@utils.track_command("ai", False, operation="init")
def ai_init_command(
    agent: Optional[str],
    location: str,
    branch: Optional[str] = None,
    overwrite: bool = False,
) -> None:
    """Install the init toolkit into the current project."""
    project_root = Path(run_context.active().run_dir)
    var = _resolve_agent(agent, project_root)

    fmt.echo("Initializing AI rules for %s from %s..." % (fmt.bold(var.name), fmt.bold(location)))

    base = _fetch_workbench_base_cli(location, branch)
    if base is None:
        return

    _install_toolkit(_INIT_TOOLKIT, base, var, project_root, overwrite=overwrite)

    status = fetch_ai_status(project_root)
    if "mcp_unavailable" in status["warnings"]:
        fmt.warning("MCP server cannot be started. Run `dlt ai status` for details.")
    if var.name == "cursor":
        fmt.warning(
            "Cursor requires you to manually enable MCP servers per-project."
            " Open Cursor Settings > MCP and enable the servers installed by dlt."
        )

    if len(load_toolkits_index()) == 1:
        fmt.echo()
        fmt.echo(
            "Now you can install your first toolkit. Use `dlt ai toolkit list` for more"
            " information."
        )


@utils.track_command("ai", False, operation="toolkit_install")
def ai_toolkit_install_command(
    name: str,
    agent: Optional[str],
    location: str,
    branch: Optional[str] = None,
    overwrite: bool = False,
    strict: bool = False,
) -> None:
    """Install toolkit components into the current project."""
    project_root = Path(run_context.active().run_dir)
    var = _resolve_agent(agent, project_root)

    fmt.echo(
        "Installing toolkit %s for %s from %s..."
        % (fmt.bold(name), fmt.bold(var.name), fmt.bold(location))
    )

    base = _fetch_workbench_base_cli(location, branch)
    if base is None:
        return

    toolkits, warnings = fetch_workbench_toolkits(base, strict=strict)
    for w in warnings:
        fmt.warning(w)
    _install_dependencies(name, toolkits, base, var, project_root)
    _install_toolkit(name, base, var, project_root, overwrite=overwrite, strict=strict)


@utils.track_command("ai", False, operation="toolkit_list")
def ai_toolkit_list_command(
    location: str,
    branch: Optional[str] = None,
) -> None:
    """List available toolkits with name and description."""
    base = _fetch_workbench_base_cli(location, branch)
    if base is None:
        return
    toolkits, warnings = fetch_workbench_toolkits(base, listed_only=True)
    for w in warnings:
        fmt.warning(w)
    if not toolkits:
        fmt.echo("No toolkits found.")
        return

    installed = load_toolkits_index()
    installed_tks: List[TToolkitInfo] = []
    available_tks: List[TToolkitInfo] = []
    for tk in toolkits.values():
        if tk["name"] in installed:
            installed_tks.append(tk)
        else:
            available_tks.append(tk)

    if installed_tks:
        fmt.echo("Installed toolkits:")
        for tk in installed_tks:
            tk_name = tk["name"]
            description = tk["description"]
            remote_version = tk["version"]
            local_version = installed[tk_name].get("version", "?")
            if remote_version != local_version:
                ver = "%s, %s available" % (
                    fmt.bold(local_version),
                    fmt.style(remote_version, fg="yellow"),
                )
            else:
                ver = fmt.bold(local_version)
            fmt.echo("  %-20s %s (%s)" % (fmt.bold(tk_name), description, ver))

    if available_tks:
        if installed_tks:
            fmt.echo("")
        fmt.echo("Available toolkits:")
        for tk in available_tks:
            tk_name = tk["name"]
            description = tk["description"]
            ver = " (%s)" % fmt.bold(tk["version"])
            fmt.echo("  %-20s %s%s" % (fmt.bold(tk_name), description, ver))


@utils.track_command("ai", False, operation="toolkit_info")
def ai_toolkit_info_command(
    name: str,
    location: str,
    branch: Optional[str] = None,
) -> None:
    """Show what's inside a toolkit."""
    info = fetch_workbench_toolkit_info(name, location, branch)
    if info is None:
        fmt.warning(
            "Toolkit %s not found (missing %s)" % (fmt.bold(name), ".claude-plugin/plugin.json")
        )
        return

    fmt.echo("Toolkit: %s" % fmt.bold(info["name"]))
    if info["description"]:
        fmt.echo("  %s" % info["description"])
    if workflow_entry_skill := info.get("workflow_entry_skill"):
        fmt.echo("Use %s skill to start!" % fmt.bold(workflow_entry_skill))

    for label, items in [
        ("Skills", info["skills"]),
        ("Commands", info["commands"]),
        ("Rules", info["rules"]),
    ]:
        if items:
            fmt.echo("\n%s:" % label)
            for item in items:
                fmt.echo("  %-20s %s" % (fmt.bold(item["name"]), item["description"]))

    if info.get("mcp_servers"):
        fmt.echo("\nMCP servers:")
        for srv_name, srv_config in info["mcp_servers"].items():
            cmd = srv_config.get("command", "")
            srv_args = " ".join(srv_config.get("args", []))
            fmt.echo("  %-20s %s %s" % (fmt.bold(srv_name), cmd, srv_args))

    if info["has_ignore"]:
        fmt.echo("\nIgnore: .claudeignore")



import contextlib
import os
import sys
import subprocess
from importlib.resources import files
import time
from typing import Any, Iterator, List
from pathlib import Path
import urllib

from dlt.common.exceptions import MissingDependencyException


# keep this, will raise if user tries to run dashboard without dependencies
try:
    import marimo
    import pyarrow
    import ibis  # noqa: I251
except ModuleNotFoundError:
    raise MissingDependencyException(
        "Workspace Dashboard",
        ["dlt[workspace]"],
        "to install the dlt workspace extra.",
    )

from dlt._workspace.helpers.dashboard.const import (
    DASHBOARD_STARTUP_TIMEOUT_S,
    DASHBOARD_STARTUP_WAIT_ON_OK_S,
    DEFAULT_DASHBOARD_PORT,
    EJECTED_APP_FILE_NAME,
    HTTP_READY_POLL_INTERVAL_S,
    HTTP_READY_TIMEOUT_S,
    HTTP_READY_WAIT_ON_OK_S,
    MARIMO_COMMAND,
    PROCESS_TERMINATE_TIMEOUT_S,
    STYLE_FILE_NAME,
)


def run_dashboard(
    pipeline_name: str = None,
    edit: bool = False,
    pipelines_dir: str = None,
    port: int = None,
    host: str = None,
    with_test_identifiers: bool = False,
    headless: bool = False,
) -> None:
    """Run dashboard blocked"""
    try:
        subprocess.run(
            run_dashboard_command(
                pipeline_name, edit, pipelines_dir, port, host, with_test_identifiers, headless
            )
        )
    except KeyboardInterrupt:
        pass


def _wait_http_up(
    url: str,
    timeout_s: float = HTTP_READY_TIMEOUT_S,
    wait_on_ok: float = HTTP_READY_WAIT_ON_OK_S,
) -> None:
    start = time.monotonic()
    while time.monotonic() - start < timeout_s:
        try:
            with urllib.request.urlopen(url, timeout=1.0):
                time.sleep(wait_on_ok)
                return
        except Exception:
            time.sleep(HTTP_READY_POLL_INTERVAL_S)
    raise TimeoutError(f"Server did not become ready: {url}")


@contextlib.contextmanager
def start_dashboard(
    pipelines_dir: str = None,
    port: int = DEFAULT_DASHBOARD_PORT,
    test_identifiers: bool = True,
    headless: bool = True,
    wait_on_ok: float = DASHBOARD_STARTUP_WAIT_ON_OK_S,
) -> Iterator[subprocess.Popen[bytes]]:
    """Launches dashboard in context manager that will kill it after use"""
    command = run_dashboard_command(
        pipeline_name=None,
        edit=False,
        pipelines_dir=pipelines_dir,
        port=port,
        with_test_identifiers=test_identifiers,
        headless=headless,
    )
    # start the dashboard process using subprocess.Popen
    proc = subprocess.Popen(command)
    try:
        _wait_http_up(
            f"http://localhost:{port}",
            timeout_s=DASHBOARD_STARTUP_TIMEOUT_S,
            wait_on_ok=wait_on_ok,
        )
        yield proc
    finally:
        proc.terminate()
        try:
            proc.wait(timeout=PROCESS_TERMINATE_TIMEOUT_S)
        except subprocess.TimeoutExpired:
            proc.kill()
            proc.wait()


def run_dashboard_command(
    pipeline_name: str = None,
    edit: bool = False,
    pipelines_dir: str = None,
    port: int = None,
    host: str = None,
    with_test_identifiers: bool = False,
    headless: bool = False,
) -> List[str]:
    """Creates cli command to run workspace dashboard"""
    from dlt._workspace.helpers.dashboard import dlt_dashboard

    ejected_app_path = os.path.join(os.getcwd(), EJECTED_APP_FILE_NAME)
    ejected_css_path = os.path.join(os.getcwd(), STYLE_FILE_NAME)
    ejected_app_exists = os.path.exists(ejected_app_path)

    # when editing, eject the app with styles to the cwd if not present already
    if edit and not ejected_app_exists:
        with open(dlt_dashboard.__file__, "r", encoding="utf-8") as f:
            app_code = f.read()
        with open(ejected_app_path, "w", encoding="utf-8") as f:
            f.write(app_code)
        css_file_path = Path(files("dlt._workspace.helpers.dashboard") / STYLE_FILE_NAME)  # type: ignore
        with open(css_file_path, "r", encoding="utf-8") as f:
            css_content = f.read()
        with open(ejected_css_path, "w", encoding="utf-8") as f:
            f.write(css_content)
        ejected_app_exists = True

    # set current pipeline
    cli_args: Any = {}
    if pipeline_name:
        cli_args["pipeline"] = pipeline_name

    # app file
    app_file_path = dlt_dashboard.__file__ if not ejected_app_exists else ejected_app_path

    dashboard_cmd = [MARIMO_COMMAND, "run" if not edit else "edit", app_file_path]

    if port:
        dashboard_cmd.append("--port")
        dashboard_cmd.append(str(port))

    if host:
        dashboard_cmd.append("--host")
        dashboard_cmd.append(host)

    if headless:
        dashboard_cmd.append("--headless")

    # collect marimo app CLI args (after the "--" separator)
    app_args: List[str] = []
    if pipeline_name:
        app_args += ["--pipeline", pipeline_name]
    if pipelines_dir:
        app_args += ["--pipelines-dir", pipelines_dir]
    if with_test_identifiers:
        app_args += ["--with_test_identifiers", "true"]
    if app_args:
        dashboard_cmd += ["--"] + app_args

    return dashboard_cmd



from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Literal,
    Optional,
    Union,
)

from dlt.common import jsonpath
from dlt.common.typing import TypedDict
from dlt.common.incremental.typing import IncrementalArgs
from dlt.extract.items import TTableHintTemplate
from dlt.extract.hints import TResourceHintsBase
from dlt.sources.helpers.rest_client.auth import AuthConfigBase, TApiKeyLocation

from dataclasses import dataclass, field

from dlt.extract.resource import DltResource

from requests import Session

from dlt.sources.helpers.rest_client.typing import HTTPMethodBasic

from dlt.sources.helpers.rest_client.paginators import (
    BasePaginator,
    JSONLinkPaginator,
    HeaderLinkPaginator,
    HeaderCursorPaginator,
    JSONResponseCursorPaginator,
    OffsetPaginator,
    PageNumberPaginator,
    SinglePagePaginator,
)

from dlt.sources.helpers.rest_client.auth import (
    HttpBasicAuth,
    BearerTokenAuth,
    APIKeyAuth,
    OAuth2ClientCredentials,
)

PaginatorType = Literal[
    "json_link",
    "json_response",  # deprecated. Use json_link instead. Will be removed in upcoming release
    "header_link",
    "header_cursor",
    "auto",
    "single_page",
    "cursor",
    "offset",
    "page_number",
]


class PaginatorTypeConfig(TypedDict, total=True):
    type: PaginatorType  # noqa


class PageNumberPaginatorConfig(PaginatorTypeConfig, total=False):
    """A paginator that uses page number-based pagination strategy."""

    base_page: Optional[int]
    page: Optional[int]
    page_param: Optional[str]
    total_path: Optional[jsonpath.TJsonPath]
    maximum_page: Optional[int]
    stop_after_empty_page: Optional[bool]
    has_more_path: Optional[jsonpath.TJsonPath]
    page_body_path: Optional[str]


class OffsetPaginatorConfig(PaginatorTypeConfig, total=False):
    """A paginator that uses offset-based pagination strategy."""

    limit: int
    offset: Optional[int]
    offset_param: Optional[str]
    offset_body_path: Optional[str]
    limit_param: Optional[str]
    limit_body_path: Optional[str]
    total_path: Optional[jsonpath.TJsonPath]
    maximum_offset: Optional[int]
    stop_after_empty_page: Optional[bool]
    has_more_path: Optional[jsonpath.TJsonPath]


class HeaderLinkPaginatorConfig(PaginatorTypeConfig, total=False):
    """A paginator that uses the 'Link' header in HTTP responses
    for pagination."""

    links_next_key: Optional[str]


class JSONLinkPaginatorConfig(PaginatorTypeConfig, total=False):
    """Locates the next page URL within the JSON response body. The key
    containing the URL can be specified using a JSON path."""

    next_url_path: Optional[jsonpath.TJsonPath]


class JSONResponseCursorPaginatorConfig(PaginatorTypeConfig, total=False):
    """Uses a cursor parameter for pagination, with the cursor value found in
    the JSON response body."""

    cursor_path: Optional[jsonpath.TJsonPath]
    cursor_param: Optional[str]
    cursor_body_path: Optional[jsonpath.TJsonPath]
    stop_after_empty_page: Optional[bool]
    has_more_path: Optional[jsonpath.TJsonPath]


class HeaderCursorPaginatorConfig(PaginatorTypeConfig, total=False):
    """A paginator that uses a cursor in the HTTP response header
    for pagination."""

    cursor_key: Optional[str]
    cursor_param: Optional[str]


PaginatorConfig = Union[
    PaginatorType,
    PageNumberPaginatorConfig,
    OffsetPaginatorConfig,
    HeaderLinkPaginatorConfig,
    HeaderCursorPaginatorConfig,
    JSONLinkPaginatorConfig,
    JSONResponseCursorPaginatorConfig,
    BasePaginator,
    SinglePagePaginator,
    HeaderLinkPaginator,
    HeaderCursorPaginator,
    JSONLinkPaginator,
    JSONResponseCursorPaginator,
    OffsetPaginator,
    PageNumberPaginator,
]


AuthType = Literal["bearer", "api_key", "http_basic", "oauth2_client_credentials"]


class AuthTypeConfig(TypedDict, total=True):
    type: AuthType  # noqa


class BearerTokenAuthConfig(TypedDict, total=False):
    """Uses `token` for Bearer authentication in "Authorization" header."""

    # we allow for a shorthand form of bearer auth, without a type
    type: Optional[AuthType]  # noqa
    token: str


class ApiKeyAuthConfig(AuthTypeConfig, total=False):
    """Uses provided `api_key` to create authorization data in the specified `location` (query, param, header, cookie) under specified `name`"""

    name: Optional[str]
    api_key: str
    location: Optional[TApiKeyLocation]


class HttpBasicAuthConfig(AuthTypeConfig, total=True):
    """Uses HTTP basic authentication"""

    username: str
    password: str


class OAuth2ClientCredentialsConfig(AuthTypeConfig, total=False):
    """Uses OAuth 2.0 client credential authorization"""

    access_token: Optional[str]
    access_token_url: str
    client_id: str
    client_secret: str
    access_token_request_data: Optional[Dict[str, Any]]
    default_token_expiration: Optional[int]
    session: Optional[Session]


# TODO: add later
# class OAuthJWTAuthConfig(AuthTypeConfig, total=True):


AuthConfig = Union[
    AuthConfigBase,
    AuthType,
    BearerTokenAuth,
    BearerTokenAuthConfig,
    APIKeyAuth,
    ApiKeyAuthConfig,
    HttpBasicAuth,
    HttpBasicAuthConfig,
    OAuth2ClientCredentials,
    OAuth2ClientCredentialsConfig,
]


class ClientConfig(TypedDict, total=False):
    base_url: str
    headers: Optional[Dict[str, str]]
    auth: Optional[AuthConfig]
    paginator: Optional[PaginatorConfig]
    session: Optional[Session]


class IncrementalRESTArgs(IncrementalArgs, total=False):
    convert: Optional[Callable[..., Any]]


class IncrementalConfig(IncrementalRESTArgs, total=False):
    start_param: Optional[str]
    end_param: Optional[str]


ParamBindType = Literal["resolve", "incremental"]


class ParamBindConfig(TypedDict):
    type: ParamBindType  # noqa


class ResolveParamConfig(ParamBindConfig):
    resource: str
    field: str


class IncrementalParamConfig(ParamBindConfig, IncrementalRESTArgs):
    pass
    # TODO: implement param type to bind incremental to
    # param_type: Optional[Literal["start_param", "end_param"]]


@dataclass
class ResolvedParam:
    param_name: str
    resolve_config: ResolveParamConfig
    field_path: jsonpath.TJsonPath = field(init=False)

    def __post_init__(self) -> None:
        self.field_path = jsonpath.compile_path(self.resolve_config["field"])


class ResponseActionDict(TypedDict, total=False):
    status_code: Optional[Union[int, str]]
    content: Optional[str]
    action: Optional[Union[str, Union[Callable[..., Any], List[Callable[..., Any]]]]]


ResponseAction = Union[ResponseActionDict, Callable[..., Any]]


class Endpoint(TypedDict, total=False):
    path: Optional[str]
    method: Optional[HTTPMethodBasic]
    params: Optional[Dict[str, Union[ResolveParamConfig, IncrementalParamConfig, Any]]]
    json: Optional[Dict[str, Any]]
    data: Optional[Any]
    paginator: Optional[PaginatorConfig]
    data_selector: Optional[jsonpath.TJsonPath]
    response_actions: Optional[List[ResponseAction]]
    incremental: Optional[IncrementalConfig]
    auth: Optional[AuthConfig]
    headers: Optional[Dict[str, Any]]


class ProcessingSteps(TypedDict, total=False):
    filter: Optional[Callable[[Any], bool]]  # noqa: A003
    map: Optional[Callable[[Any], Any]]  # noqa: A003
    yield_map: Optional[Callable[[Any], Iterator[Any]]]  # noqa: A003


class ResourceBase(TResourceHintsBase, total=False):
    """Defines hints that may be passed to `dlt.resource` decorator"""

    max_table_nesting: Optional[int]
    selected: Optional[bool]
    parallelized: Optional[bool]
    processing_steps: Optional[List[ProcessingSteps]]


class EndpointResourceBase(ResourceBase, total=False):
    endpoint: Optional[Union[str, Endpoint]]
    include_from_parent: Optional[List[str]]


class EndpointResource(EndpointResourceBase, total=False):
    name: TTableHintTemplate[str]


class RESTAPIConfigBase(TypedDict):
    client: ClientConfig
    resources: List[Union[str, EndpointResource, DltResource]]


class RESTAPIConfig(RESTAPIConfigBase, total=False):
    resource_defaults: Optional[EndpointResourceBase]



import os
from typing import Dict, List

from dlt.common.data_writers import TDataItemFormat, DataWriter, FileWriterSpec
from dlt.common.metrics import DataWriterMetrics
from dlt.common.schema import Schema
from dlt.common.storages import (
    NormalizeStorageConfiguration,
    NormalizeStorage,
    DataItemStorage,
    FileStorage,
    PackageStorage,
    LoadPackageInfo,
    create_load_id,
)
from dlt.common.storages.exceptions import LoadPackageNotFound
from dlt.common.utils import uniq_id


class ExtractorItemStorage(DataItemStorage):
    def __init__(self, package_storage: PackageStorage, writer_spec: FileWriterSpec) -> None:
        """Data item storage using `storage` to manage load packages"""
        super().__init__(writer_spec)
        self.package_storage = package_storage

    def _get_data_item_path_template(self, load_id: str, _: str, table_name: str) -> str:
        file_name = PackageStorage.build_job_file_name(table_name, "%s")
        file_path = self.package_storage.get_job_file_path(
            load_id, PackageStorage.NEW_JOBS_FOLDER, file_name
        )
        return self.package_storage.storage.make_full_path(file_path)


class ExtractStorage(NormalizeStorage):
    """Wrapper around multiple extractor storages with different file formats"""

    def __init__(self, config: NormalizeStorageConfiguration) -> None:
        super().__init__(True, config)
        # always create new packages in a unique folder for each instance so
        # extracts are isolated ie. if they fail
        self.new_packages_folder = uniq_id(8)
        self.storage.create_folder(self.new_packages_folder, exists_ok=True)
        self.new_packages = PackageStorage(
            FileStorage(os.path.join(self.storage.storage_path, self.new_packages_folder)), "new"
        )
        self.item_storages: Dict[TDataItemFormat, ExtractorItemStorage] = {
            "object": ExtractorItemStorage(
                self.new_packages, DataWriter.writer_spec_from_file_format("typed-jsonl", "object")
            ),
            "arrow": ExtractorItemStorage(
                self.new_packages, DataWriter.writer_spec_from_file_format("parquet", "arrow")
            ),
            "model": ExtractorItemStorage(
                self.new_packages, DataWriter.writer_spec_from_file_format("model", "model")
            ),
        }

    def create_load_package(self, schema: Schema, reuse_exiting_package: bool = True) -> str:
        """Creates a new load package for given `schema` or returns if such package already exists.

        You can prevent reuse of the existing package by setting `reuse_exiting_package` to False
        """
        load_id: str = None
        if reuse_exiting_package:
            # look for existing package with the same schema name
            # TODO: we may cache this mapping but fallback to files is required if pipeline restarts
            load_ids = self.new_packages.list_packages()
            for load_id in load_ids:
                if self.new_packages.schema_name(load_id) == schema.name:
                    break
                load_id = None
        if not load_id:
            load_id = create_load_id()
            self.new_packages.create_package(load_id)
        # always save schema
        self.new_packages.save_schema(load_id, schema)
        return load_id

    def close_writers(self, load_id: str, skip_flush: bool = False) -> None:
        for storage in self.item_storages.values():
            storage.close_writers(load_id, skip_flush=skip_flush)

    def closed_files(self, load_id: str) -> List[DataWriterMetrics]:
        files = []
        for storage in self.item_storages.values():
            files.extend(storage.closed_files(load_id))
        return files

    def remove_closed_files(self, load_id: str) -> None:
        for storage in self.item_storages.values():
            storage.remove_closed_files(load_id)

    def commit_new_load_package(self, load_id: str, schema: Schema) -> None:
        self.new_packages.save_schema(load_id, schema)
        self.storage.rename_tree(
            os.path.join(self.new_packages_folder, self.new_packages.get_package_path(load_id)),
            os.path.join(
                NormalizeStorage.EXTRACTED_FOLDER, self.new_packages.get_package_path(load_id)
            ),
        )

    def delete_empty_extract_folder(self) -> None:
        """Deletes temporary extract folder if empty"""
        self.storage.delete_folder(self.new_packages_folder, recursively=False)

    def get_load_package_info(self, load_id: str) -> LoadPackageInfo:
        """Returns information on temp and extracted packages"""
        try:
            return self.new_packages.get_load_package_info(load_id)
        except LoadPackageNotFound:
            return self.extracted_packages.get_load_package_info(load_id)



import io
import os
import contextlib
import sys
import multiprocessing
import platform
from typing import Optional

from dlt.common.runtime.typing import TExecutionContext, TVersion, TExecInfoNames
from dlt.common.typing import StrStr, StrAny, List
from dlt.common.utils import filter_env_vars
from dlt.version import __version__, DLT_PKG_NAME


# if one of these environment variables is set, we assume to be running in CI env
CI_ENVIRONMENT_TELL = [
    "bamboo.buildKey",
    "BUILD_ID",
    "BUILD_NUMBER",
    "BUILDKITE",
    "CI",
    "CIRCLECI",
    "CONTINUOUS_INTEGRATION",
    "GITHUB_ACTIONS",
    "HUDSON_URL",
    "JENKINS_URL",
    "TEAMCITY_VERSION",
    "TRAVIS",
    "CODEBUILD_BUILD_ARN",
    "CODEBUILD_BUILD_ID",
    "CODEBUILD_BATCH_BUILD_IDENTIFIER",
]


def exec_info_names() -> List[TExecInfoNames]:
    """Get names of execution environments"""
    names: List[TExecInfoNames] = []
    if kube_pod_info():
        names.append("kubernetes")
    if is_docker():
        names.append("docker")
    if is_codespaces():
        names.append("codespaces")
    if is_github_actions():
        names.append("github_actions")
    if is_notebook():
        names.append("notebook")
    if is_colab():
        names.append("colab")
    if is_running_in_airflow_task():
        names.append("airflow")
    if is_running_in_dagster_task():
        names.append("dagster")
    if is_running_in_prefect_flow():
        names.append("prefect")
    if is_marimo():
        names.append("marimo")
    if is_aws_lambda():
        names.append("aws_lambda")
    if is_gcp_cloud_function():
        names.append("gcp_cloud_function")
    if is_streamlit():
        names.append("streamlit")
    if is_claude_code():
        names.append("claude_code")
    if is_claude_code_cli():
        names.append("claude_code_cli")
    if is_cursor():
        names.append("cursor")
    if is_codex():
        names.append("codex")
    return names


def is_codespaces() -> bool:
    return "CODESPACES" in os.environ


def is_github_actions() -> bool:
    return "GITHUB_ACTIONS" in os.environ


def is_streamlit() -> bool:
    return "STREAMLIT_SERVER_PORT" in os.environ


def is_claude_code() -> bool:
    """Return True if running under Claude Code (CLI or API).

    Detects CLAUDECODE environment variable set by Claude Code.
    https://code.claude.com/docs/en/settings
    """
    return os.environ.get("CLAUDECODE") == "1"


def is_claude_code_cli() -> bool:
    """Return True if running under Claude Code CLI specifically.

    Detects CLAUDECODE=1 and CLAUDE_CODE_ENTRYPOINT=cli environment variables.
    Note: Both claude_code and claude_code_cli will be present when running via CLI.
    https://code.claude.com/docs/en/settings
    """
    return os.environ.get("CLAUDECODE") == "1" and os.environ.get("CLAUDE_CODE_ENTRYPOINT") == "cli"


def is_cursor() -> bool:
    """Return True if running in Cursor IDE Agent.

    Detects CURSOR_AGENT environment variable set when Cursor Agent is running.
    https://cursor.com/docs/agent/terminal
    """
    return os.environ.get("CURSOR_AGENT") == "1"


def is_codex() -> bool:
    """Return True if running in OpenAI Codex.

    Detects CODEX_CI, CODEX_THREAD_ID, or CODEX_SANDBOX environment variables.
    https://developers.openai.com/codex/cli/features/
    """
    return (
        os.environ.get("CODEX_CI") == "1"
        or os.environ.get("CODEX_THREAD_ID") is not None
        or os.environ.get("CODEX_SANDBOX") is not None
    )


def is_notebook() -> bool:
    try:
        return bool(str(get_ipython()))  # type: ignore
    except NameError:
        return False


def is_pyodide() -> bool:
    return sys.platform == "emscripten"


def platform_supports_threading() -> bool:
    return not is_pyodide()


def is_colab() -> bool:
    try:
        return "COLAB_RELEASE_TAG" in os.environ or "google.colab" in str(get_ipython())  # type: ignore
    except NameError:
        return False


def airflow_info() -> StrAny:
    try:
        with contextlib.redirect_stdout(io.StringIO()), contextlib.redirect_stderr(io.StringIO()):
            from airflow.operators.python import get_current_context

            get_current_context()
            return {"AIRFLOW_TASK": True}
    except Exception:
        return None


def is_airflow_installed() -> bool:
    try:
        with contextlib.redirect_stdout(io.StringIO()), contextlib.redirect_stderr(io.StringIO()):
            import airflow
        return True
    except Exception:
        return False


def is_running_in_airflow_task() -> bool:
    return "AIRFLOW_CTX_TASK_ID" in os.environ


def is_running_in_dagster_task() -> bool:
    # module must be imported in dagster task
    if dagster_ := sys.modules.get("dagster._core.execution.context.compute"):
        try:
            return dagster_.current_execution_context.get() is not None
        except Exception:
            pass
    return False


def is_running_in_prefect_flow() -> bool:
    # check if prefect module is imported must be the case if running in flow
    if pf := sys.modules.get("prefect"):
        get_ctx = getattr(getattr(pf, "context", None), "get_run_context", None)
        if callable(get_ctx):
            try:
                get_ctx()
                return True
            except Exception:
                pass
    return False


def is_marimo() -> bool:
    if marimo_ := sys.modules.get("marimo"):
        get_ctx = getattr(marimo_, "running_in_notebook", None)
        if callable(get_ctx):
            try:
                return get_ctx()  # type: ignore[no-any-return]
            except Exception:
                pass
    return False


def dlt_version_info(pipeline_name: Optional[str]) -> StrStr:
    """Gets dlt version info including commit and image version available in docker"""
    version_info = {"dlt_version": __version__}
    if pipeline_name:
        version_info["pipeline_name"] = pipeline_name
    # extract envs with build info
    version_info.update(filter_env_vars(["COMMIT_SHA", "IMAGE_VERSION"]))

    return version_info


def kube_pod_info() -> StrStr:
    """Extracts information on pod name, namespace and node name if running on Kubernetes"""
    return filter_env_vars(["KUBE_NODE_NAME", "KUBE_POD_NAME", "KUBE_POD_NAMESPACE"])


def github_info() -> StrStr:
    """Extracts github info"""
    info = filter_env_vars(["GITHUB_USER", "GITHUB_REPOSITORY", "GITHUB_REPOSITORY_OWNER"])
    # set GITHUB_REPOSITORY_OWNER as github user if not present. GITHUB_REPOSITORY_OWNER is available in github action context
    if "github_user" not in info and "github_repository_owner" in info:
        info["github_user"] = info["github_repository_owner"]  # type: ignore
    return info


def in_continuous_integration() -> bool:
    """Returns `True` if currently running inside a continuous integration context."""
    return any(env in os.environ for env in CI_ENVIRONMENT_TELL)


def is_docker() -> bool:
    """Guess if we are running in docker environment.

    https://stackoverflow.com/questions/20010199/how-to-determine-if-a-process-runs-inside-lxc-docker

    Returns:
        `True` if we are running inside docker, `False` otherwise.
    """
    # first we try to use the env
    try:
        os.stat("/.dockerenv")
        return True
    except Exception:
        pass

    # if that didn't work, try to use proc information
    try:
        with open("/proc/self/cgroup", mode="r", encoding="utf-8") as f:
            return "docker" in f.read()
    except Exception:
        return False


def is_aws_lambda() -> bool:
    "Return True if the process is running in the serverless platform AWS Lambda"
    return os.environ.get("AWS_LAMBDA_FUNCTION_NAME") is not None


def is_gcp_cloud_function() -> bool:
    "Return True if the process is running in the serverless platform GCP Cloud Functions"
    return os.environ.get("FUNCTION_NAME") is not None


def get_plus_version() -> TVersion:
    "Gets dlthub library version"
    try:
        from dlthub.version import __version__, PKG_NAME

        return TVersion(name=PKG_NAME, version=__version__)
    except Exception:
        return None


def run_context_name() -> str:
    try:
        from dlt.common.configuration.container import Container
        from dlt.common.configuration.specs.pluggable_run_context import PluggableRunContext

        container = Container()
        if PluggableRunContext in container:
            return container[PluggableRunContext].context.name

    except Exception:
        pass

    return "dlt"


def get_execution_context() -> TExecutionContext:
    "Get execution context information"
    context = TExecutionContext(
        ci_run=in_continuous_integration(),
        python=sys.version.split(" ")[0],
        cpu=multiprocessing.cpu_count(),
        exec_info=exec_info_names(),
        os=TVersion(name=platform.system(), version=platform.release()),
        library=TVersion(name=DLT_PKG_NAME, version=__version__),
        run_context=run_context_name(),
    )
    if plus_version := get_plus_version():
        context["dlthub"] = plus_version

    return context



# mypy: disable-error-code="return-value, no-any-return"
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional

import sqlglot
import sqlglot.expressions as sge
from pydantic import Field
from fastmcp.exceptions import ToolError

import dlt
from dlt.common.configuration.specs.pluggable_run_context import ProfilesRunContext
from dlt.common.runtime.run_context import active as active_run_context
from dlt.common.schema.exceptions import IncompatibleSchemaException
from dlt.common.schema.schema import Schema
from dlt.common.typing import Annotated
from dlt._workspace.cli import formatters
from dlt._workspace.cli.utils import (
    fetch_profiles_list,
    fetch_schema_export,
    fetch_workspace_info,
    list_local_pipelines,
)
from dlt._workspace.mcp.context import with_mcp_tool_telemetry

TResultFormat = Literal["markdown", "jsonl"]


def _attach(pipeline_name: str) -> dlt.Pipeline:
    """Attach to a pipeline by name."""
    return dlt.attach(pipeline_name)


def _get_unified_schema(pipeline: dlt.Pipeline) -> Schema:
    """Build a unified schema from all pipeline schemas.

    Falls back to default schema if naming conventions are incompatible.
    """
    schema_names = list(pipeline.schemas)
    if len(schema_names) <= 1:
        return pipeline.default_schema
    default = pipeline.default_schema
    others = [pipeline.schemas[n] for n in schema_names if n != default.name]
    try:
        return default.unify_schemas(others)
    except IncompatibleSchemaException:
        return default


def _get_dataset(pipeline_name: str) -> dlt.Dataset:
    """Attach to a pipeline and return its dataset with unified schema."""
    pipeline = _attach(pipeline_name)
    return pipeline.dataset(schema=_get_unified_schema(pipeline))


@with_mcp_tool_telemetry()
def list_pipelines() -> List[str]:
    """List all dlt pipelines available in this workspace"""
    ctx = active_run_context()
    # in OSS context (no profiles), filter to pipelines created in this project
    project_dir = None if isinstance(ctx, ProfilesRunContext) else ctx.local_dir
    _, pipelines = list_local_pipelines(sort_by_trace=False, run_dir=project_dir)
    return [p["name"] for p in pipelines]


@with_mcp_tool_telemetry()
def get_workspace_info() -> Dict[str, Any]:
    """Get workspace info: name, directories, active profile, and config providers."""

    info = fetch_workspace_info()
    # prune bulky file-tracking data from toolkit entries
    for entry in info.get("installed_toolkits", {}).values():
        entry.pop("files", None)
    return info


@with_mcp_tool_telemetry()
def list_profiles() -> List[Dict[str, Any]]:
    """List all available workspace profiles with status flags (current, pinned, configured)."""
    return fetch_profiles_list()  # type: ignore[return-value]


@with_mcp_tool_telemetry()
def list_tables(
    pipeline_name: str,
) -> Dict[str, Any]:
    """List all data tables for a pipeline."""
    pipeline = _attach(pipeline_name)
    schema = _get_unified_schema(pipeline)
    return {"tables": [table["name"] for table in schema.data_tables()]}


@with_mcp_tool_telemetry()
def get_table_schema(
    pipeline_name: str,
    table_name: str,
) -> Dict[str, Any]:
    """Get table schema with column names, data types, and escaped sql_identifier fields."""
    try:
        dataset = _get_dataset(pipeline_name)
        schema = dataset.schema
        table_dict: Dict[str, Any] = schema.get_table(table_name)  # type: ignore[assignment]

        dialect = dataset.sql_client.capabilities.sqlglot_dialect or "duckdb"
        table_dict["sql_dialect"] = dialect

        table_dict["sql_identifier"] = dataset.sql_client.escape_column_name(
            schema.naming.normalize_tables_path(table_dict["name"])
        )

        for col_schema in table_dict["columns"].values():
            col_schema["sql_identifier"] = dataset.sql_client.escape_column_name(
                schema.naming.normalize_tables_path(col_schema["name"])
            )

        stored_schema = schema.to_dict(remove_defaults=True, bump_version=False)
        return stored_schema["tables"][table_name]
    except Exception as e:
        raise ToolError(
            "Tool `get_table_schema` failed. Verify `pipeline_name` and `table_name`. "
            "If the error persists, try starting a new conversation."
        ) from e


@with_mcp_tool_telemetry()
def get_table_create_sql(
    pipeline_name: str,
    table_name: str,
) -> str:
    """Get CREATE TABLE DDL for the table in the destination's SQL dialect."""
    from dlt.common.libs.sqlglot import to_sqlglot_type

    try:
        dataset = _get_dataset(pipeline_name)
        schema = dataset.schema
        table_dict = schema.get_table(table_name)
        dialect = dataset.sql_client.capabilities.sqlglot_dialect or "duckdb"

        col_defs: List[sge.ColumnDef] = []
        for col in table_dict["columns"].values():
            col_name = dataset.sql_client.escape_column_name(
                schema.naming.normalize_tables_path(col["name"]),
                quote=False,
            )
            sg_type = to_sqlglot_type(
                col["data_type"],
                precision=col.get("precision"),
                scale=col.get("scale"),
                timezone=col.get("timezone"),
                nullable=col.get("nullable", True),
            )
            col_def = sge.ColumnDef(
                this=sge.to_identifier(col_name, quoted=True),
                kind=sg_type,
            )
            col_desc = col.get("description")
            if col_desc:
                col_def.args["constraints"] = [
                    sge.ColumnConstraint(
                        kind=sge.CommentColumnConstraint(this=sge.Literal.string(col_desc))
                    )
                ]
            col_defs.append(col_def)

        table_id = dataset.sql_client.escape_column_name(
            schema.naming.normalize_tables_path(table_dict["name"]),
            quote=False,
        )
        create = sge.Create(
            this=sge.Schema(
                this=sge.to_identifier(table_id, quoted=True),
                expressions=col_defs,
            ),
            kind="TABLE",
        )
        table_desc = table_dict.get("description")
        if table_desc:
            create.args["properties"] = sge.Properties(
                expressions=[sge.SchemaCommentProperty(this=sge.Literal.string(table_desc))]
            )
        return create.sql(dialect=dialect, pretty=True)
    except Exception as e:
        raise ToolError(
            "Tool `get_table_create_sql` failed. Verify `pipeline_name` and `table_name`. "
            "If the error persists, try starting a new conversation."
        ) from e


@with_mcp_tool_telemetry()
def preview_table(
    pipeline_name: str,
    table_name: str,
    output_format: Annotated[
        TResultFormat,
        Field(description="Output format: 'markdown' or 'jsonl'"),
    ] = "markdown",
) -> str:
    """Get the first 10 rows from a table."""
    try:
        relation = _get_dataset(pipeline_name)[table_name].limit(10)
        columns = relation.columns
        rows = relation.fetchall()
        if output_format == "jsonl":
            return formatters.render_jsonl(columns, rows)
        return formatters.render_md_table(columns, rows)
    except Exception as e:
        raise ToolError(
            "Tool `preview_table` failed. Verify `pipeline_name` and `table_name`. "
            "If the error persists, try starting a new conversation."
        ) from e


@with_mcp_tool_telemetry()
def execute_sql_query(
    pipeline_name: str,
    sql_select_query: Annotated[
        str,
        Field(description="SQL SELECT query to execute (only SELECT is allowed)"),
    ],
    output_format: Annotated[
        TResultFormat,
        Field(description="Output format: 'markdown' or 'jsonl'"),
    ] = "markdown",
) -> str:
    """Execute a read-only SQL query against the pipeline's destination dataset."""
    parsed = sqlglot.parse(sql_select_query)
    if any(
        isinstance(expr, (sqlglot.exp.Insert, sqlglot.exp.Update, sqlglot.exp.Delete))
        for expr in parsed
    ):
        raise ToolError("Data modification statements are not allowed")

    dataset = _get_dataset(pipeline_name)
    relation = dataset(sql_select_query)
    columns = relation.columns
    rows = relation.fetchall()
    if output_format == "jsonl":
        return formatters.render_jsonl(columns, rows)
    return formatters.render_md_table(columns, rows)


@with_mcp_tool_telemetry()
def get_row_counts(
    pipeline_name: str,
    output_format: Annotated[
        TResultFormat,
        Field(description="Output format: 'markdown' or 'jsonl'"),
    ] = "markdown",
) -> str:
    """Get row counts for all data tables in a pipeline."""
    try:
        dataset = _get_dataset(pipeline_name)
        relation = dataset.row_counts()
        columns = relation.columns
        rows = relation.fetchall()
        if output_format == "jsonl":
            return formatters.render_jsonl(columns, rows)
        return formatters.render_md_table(columns, rows)
    except Exception as e:
        raise ToolError(
            "Tool `get_row_counts` failed. Verify `pipeline_name`. "
            "If the error persists, try starting a new conversation."
        ) from e


TMcpSchemaFormat = Literal["mermaid", "yaml", "dbml"]


@with_mcp_tool_telemetry()
def export_schema(
    pipeline_name: str,
    schema_name: Optional[str] = None,
    hide_columns: Annotated[
        bool,
        Field(description="Hide column details for better readability of large schemas"),
    ] = False,
    output_format: Annotated[
        TMcpSchemaFormat,
        Field(description="Output format: 'mermaid', 'yaml', or 'dbml'"),
    ] = "mermaid",
    save_to_file: Annotated[
        Optional[str],
        Field(
            description=(
                "Save the schema to this file path instead of returning it."
                " Use an absolute path (e.g. /home/user/schema.yaml)."
            )
        ),
    ] = None,
) -> str:
    """Export the pipeline schema as a diagram or structured dump."""
    try:
        pipeline = _attach(pipeline_name)
        schema = pipeline.schemas[schema_name] if schema_name else _get_unified_schema(pipeline)
        export = fetch_schema_export(schema, format_=output_format, hide_columns=hide_columns)
        if save_to_file:
            path = Path(save_to_file)
            path.parent.mkdir(parents=True, exist_ok=True)
            path.write_text(export["content"], encoding="utf-8")
            return "Schema saved to %s" % path
        return export["content"]
    except Exception as e:
        raise ToolError(
            "Tool `export_schema` failed. Verify `pipeline_name`. "
            "If the error persists, try starting a new conversation."
        ) from e


@with_mcp_tool_telemetry()
def get_local_pipeline_state(
    pipeline_name: str,
) -> Dict[str, Any]:
    """Get pipeline state: incremental cursors, resource state, and source state."""
    try:
        pipeline = _attach(pipeline_name)
        return pipeline.state  # type: ignore[return-value]
    except Exception as e:
        raise ToolError(
            "Tool `get_local_pipeline_state` failed. Verify `pipeline_name`. "
            "If the error persists, try starting a new conversation."
        ) from e


@with_mcp_tool_telemetry()
def pipeline_trace(
    pipeline_name: str,
) -> Dict[str, Any]:
    """Get the trace of the last pipeline run: timing, step outcomes, and errors."""
    pipeline = _attach(pipeline_name)
    trace = pipeline.last_trace
    if trace is None:
        raise ToolError(
            f"No trace found for pipeline '{pipeline_name}'. "
            "The pipeline may not have been run yet."
        )
    return trace.asdict()


__tools__ = (
    list_pipelines,
    list_profiles,
    list_tables,
    get_table_schema,
    get_table_create_sql,
    preview_table,
    execute_sql_query,
    get_row_counts,
    export_schema,
    get_local_pipeline_state,
    # TODO: pipeline_trace returns raw trace.asdict() which is too large for MCP.
    #   Return a summary (timing, step outcomes, errors) like the CLI trace command does.
)



import dataclasses
from typing import Dict, Literal, Optional, Final
from typing_extensions import Annotated
from urllib.parse import urlparse

from dlt.common.configuration import configspec, NotResolved
from dlt.common.configuration.specs.base_configuration import CredentialsConfiguration
from dlt.common.destination.client import DestinationClientDwhConfiguration
from dlt.common.utils import digest128

TWeaviateBatchConsistency = Literal["ONE", "QUORUM", "ALL"]
TWeaviateConnectionType = Literal["cloud", "local", "custom"]


@configspec
class WeaviateCredentials(CredentialsConfiguration):
    url: str = "http://localhost:8080"
    api_key: Optional[str] = None
    additional_headers: Optional[Dict[str, str]] = None
    # Optional ports for custom/local connections
    http_port: Optional[int] = None
    grpc_port: Optional[int] = None

    def __str__(self) -> str:
        """Used to display user friendly data location"""
        # assuming no password in url scheme for Weaviate
        return self.url


@configspec
class WeaviateClientConfiguration(DestinationClientDwhConfiguration):
    destination_type: Final[str] = dataclasses.field(default="weaviate", init=False, repr=False, compare=False)  # type: ignore
    # make it optional so empty dataset is allowed
    dataset_name: Annotated[Optional[str], NotResolved()] = dataclasses.field(
        default=None, init=False, repr=False, compare=False
    )

    batch_size: int = 100
    batch_workers: int = 1
    batch_consistency: TWeaviateBatchConsistency = "ONE"
    batch_retries: int = 5

    conn_timeout: float = 10.0
    read_timeout: float = 3 * 60.0
    startup_period: int = 5

    dataset_separator: str = "_"

    # Connection type: "cloud" for Weaviate Cloud, "local" for Docker, "custom" for self-hosted
    # If None, auto-detected from URL pattern
    connection_type: Optional[TWeaviateConnectionType] = None

    credentials: WeaviateCredentials = None
    vectorizer: str = "text2vec-openai"
    module_config: Dict[str, Dict[str, str]] = dataclasses.field(
        default_factory=lambda: {
            "text2vec-openai": {
                "model": "ada",
                "modelVersion": "002",
                "type": "text",
            }
        }
    )

    def fingerprint(self) -> str:
        """Returns a fingerprint of host part of a connection string"""

        if self.credentials and self.credentials.url:
            hostname = urlparse(self.credentials.url).hostname
            return digest128(hostname)
        return ""



import os
import sys
import logging
import time
import warnings

from collections import defaultdict
from typing import (
    Any,
    ContextManager,
    Dict,
    Literal,
    TYPE_CHECKING,
    DefaultDict,
    NamedTuple,
    Optional,
    Union,
    TextIO,
)

if TYPE_CHECKING:
    from tqdm import tqdm
    import enlighten
    from enlighten import Counter as EnlCounter, StatusBar as EnlStatusBar, Manager as EnlManager
    from alive_progress import alive_bar
else:
    tqdm = EnlCounter = EnlStatusBar = EnlManager = Any

from dlt.common import logger as dlt_logger
from dlt.common.time import MonotonicPreciseTime
from dlt.common.exceptions import MissingDependencyException
from dlt.common.runtime.collector_base import Collector
from dlt.common.warnings import DltDeprecationWarning


class NullCollector(Collector):
    """A default counter that does not count anything."""

    def update(
        self,
        name: str,
        inc: int = 1,
        total: int = None,
        inc_total: int = None,
        message: str = None,
        label: str = None,
    ) -> None:
        pass

    def _start(self, step: str) -> None:
        pass

    def _stop(self) -> None:
        pass


class DictCollector(Collector):
    """A collector that just counts"""

    def __init__(self) -> None:
        self.step = None
        self.counters: DefaultDict[str, int] = None

    def update(
        self,
        name: str,
        inc: int = 1,
        total: int = None,
        inc_total: int = None,
        message: str = None,
        label: str = None,
    ) -> None:
        assert not label, "labels not supported in dict collector"
        self.counters[name] += inc

    def _start(self, step: str) -> None:
        self.counters = defaultdict(int)

    def _stop(self) -> None:
        self.counters = None


class LogCollector(Collector):
    """A Collector that shows progress by writing to a Python logger or a console"""

    logger: Union[logging.Logger, TextIO, Literal["stdout", "dlt_logger"]]
    log_level: int

    class CounterInfo(NamedTuple):
        description: str
        start_time: float
        total: Optional[int]

    def __init__(
        self,
        log_period: float = 1.0,
        logger: Union[logging.Logger, TextIO, Literal["stdout", "dlt_logger"]] = "stdout",
        log_level: int = logging.INFO,
        dump_system_stats: bool = True,
    ) -> None:
        """
        Collector writing to a `logger` every `log_period` seconds.

        Args:
            log_period (float, optional): Time period in seconds between log updates. Defaults to 1.0.
            logger: Logger or text stream to write log messages to. Use ``"stdout"`` (default) to
                write to the current ``sys.stdout`` at call time, ``"dlt_logger"`` to use the
                internal dlt logger, a ``logging.Logger`` instance, or any ``TextIO`` stream.
            log_level (str, optional): Log level for the logger. Defaults to INFO level
            dump_system_stats (bool, optional): Log memory and cpu usage. Defaults to True
        """
        self.step = None
        self.log_period = log_period
        if logger is None:
            warnings.warn(
                DltDeprecationWarning(
                    "Passing logger=None to LogCollector is deprecated."
                    " Use logger='dlt_logger' instead",
                    since="1.22.0",
                ),
                stacklevel=2,
            )
            logger = "dlt_logger"
        self.logger = logger
        self.log_level = log_level
        self.counters: DefaultDict[str, int] = None
        self.counter_info: Dict[str, LogCollector.CounterInfo] = None
        self.messages: Dict[str, Optional[str]] = None
        if dump_system_stats:
            try:
                import psutil  # noqa: F401
            except ImportError:
                warnings.warn(
                    "psutil dependency is not installed and memory stats will not be available."
                    " Add psutil to your environment or pass dump_system_stats=False to disable"
                    " this warning.",
                    stacklevel=2,
                )
                dump_system_stats = False
        self.dump_system_stats = dump_system_stats
        self._clock = MonotonicPreciseTime()
        self.last_log_time: float = None

    def update(
        self,
        name: str,
        inc: int = 1,
        total: int = None,
        inc_total: int = None,
        message: str = None,
        label: str = None,
    ) -> None:
        counter_key = f"{name}_{label}" if label else name

        if counter_key not in self.counters:
            self.counters[counter_key] = 0
            self.counter_info[counter_key] = LogCollector.CounterInfo(
                description=f"{name} ({label})" if label else name,
                start_time=self._clock(),
                total=total,
            )
            self.messages[counter_key] = None
            self.last_log_time = None
        else:
            counter_info = self.counter_info[counter_key]
            if inc_total:
                self.counter_info[counter_key] = LogCollector.CounterInfo(
                    description=counter_info.description,
                    start_time=counter_info.start_time,
                    total=counter_info.total + inc_total,
                )

        self.counters[counter_key] += inc
        if message is not None:
            self.messages[counter_key] = message
        self.maybe_log()

    def maybe_log(self) -> None:
        """Check if should report and if so, call self.on_log"""
        current_time = self._clock()
        if self.last_log_time is None or current_time - self.last_log_time >= self.log_period:
            self.on_log()
            self.last_log_time = self._clock()

    def _counter_to_log_line(
        self, counter_key: str, count: int, info: CounterInfo, current_time: float
    ) -> str:
        """
        Convert a single counter to a log line.
        Example:
            Resources: 0/1 (0.0%) | Time: 0.01s | Rate: 0.00/s
        """
        elapsed_time = current_time - info.start_time
        items_per_second = (count / elapsed_time) if elapsed_time > 0 else 0

        progress = f"{count}/{info.total}" if info.total else f"{count}"
        percentage = f"({count / info.total * 100:.1f}%)" if info.total else ""
        elapsed_time_str = f"{elapsed_time:.2f}s"
        items_per_second_str = f"{items_per_second:.2f}/s"
        message = (
            f"[{self.messages[counter_key]}]" if self.messages[counter_key] is not None else ""
        )

        return (
            f"{info.description}: {progress} {percentage} | Time: {elapsed_time_str} | Rate:"
            f" {items_per_second_str} {message}"
        ).strip()

    def _system_stats_to_log_line(self) -> str:
        """Convert system stats to a log line format."""
        try:
            import psutil

            process = psutil.Process(os.getpid())
            mem_info = process.memory_info()
            current_mem = mem_info.rss / (1024**2)  # Convert to MB
            mem_percent = psutil.virtual_memory().percent
            cpu_percent = process.cpu_percent()
            return (
                f"Memory usage: {current_mem:.2f} MB ({mem_percent:.2f}%) | CPU usage:"
                f" {cpu_percent:.2f}%"
            )
        except ImportError:
            return "System stats unavailable (psutil not installed)"

    def dump_counters(self) -> None:
        """Dump all counters to log using the shared formatting methods."""
        current_time = self._clock()
        log_lines = []

        step_header = f" {self.step} ".center(80, "-")
        log_lines.append(step_header)

        for name, count in self.counters.items():
            info = self.counter_info[name]
            log_lines.append(self._counter_to_log_line(name, count, info, current_time))

        if self.dump_system_stats:
            log_lines.append(self._system_stats_to_log_line())

        log_lines.append("")
        log_message = "\n".join(log_lines)
        if self.logger == "dlt_logger":
            self.logger = dlt_logger.LOGGER or "stdout"
        self._log(self.log_level, log_message)

    def _log(self, log_level: int, log_message: str) -> None:
        if isinstance(self.logger, (logging.Logger, logging.LoggerAdapter)):
            self.logger.log(log_level, log_message)
        else:
            dest = sys.stdout if isinstance(self.logger, str) else self.logger
            dest.write(log_message + "\n")

    def _start(self, step: str) -> None:
        self.counters = defaultdict(int)
        self.counter_info = {}
        self.messages = {}
        self.last_log_time = self._clock()

    def _stop(self) -> None:
        self.on_log()
        self.counters = None
        self.counter_info = None
        self.messages = None
        self.last_log_time = None

    def on_log(self) -> None:
        self.dump_counters()


class TqdmCollector(Collector):
    """A Collector that shows progress with `tqdm` progress bars"""

    def __init__(self, single_bar: bool = False, **tqdm_kwargs: Any) -> None:
        """A Collector that uses tqdm to display counters as progress bars. Set `single_bar` to True to show just the main progress bar. Pass any config to tqdm in kwargs"""
        try:
            global tqdm
            from tqdm import tqdm
        except ModuleNotFoundError:
            raise MissingDependencyException(
                "TqdmCollector", ["tqdm"], "We need tqdm to display progress bars."
            )
        self.step = None
        self.single_bar = single_bar
        self._bars: Dict[str, tqdm[None]] = {}
        self.tqdm_kwargs = tqdm_kwargs or {}

    def update(
        self,
        name: str,
        inc: int = 1,
        total: int = None,
        inc_total: int = None,
        message: str = None,
        label: str = "",
    ) -> None:
        key = f"{name}_{label}"
        bar = self._bars.get(key)
        if bar is None:
            if label:
                name = f"{name}[{label}]"
            if len(self._bars) == 0:
                desc = self.step + ": " + name
            else:
                # do not add any more counters
                if self.single_bar:
                    return
                desc = name
            bar = tqdm(desc=desc, total=total, leave=False, **self.tqdm_kwargs)
            bar.refresh()
            self._bars[key] = bar
        else:
            if inc_total:
                bar.total += inc_total
                bar.refresh()
        if message:
            bar.set_postfix_str(message)
        bar.update(inc)

    def _start(self, step: str) -> None:
        self._bars = {}

    def _stop(self) -> None:
        for bar in self._bars.values():
            bar.refresh()
            bar.close()
        self._bars.clear()


class AliveCollector(Collector):
    """A Collector that shows progress with `alive-progress` progress bars"""

    def __init__(self, single_bar: bool = True, **alive_kwargs: Any) -> None:
        """Collector that uses alive_progress to display counters as progress bars. Set `single_bar` to True to show just the main progress bar. Pass any config to alive_progress in kwargs"""
        try:
            global alive_bar
            from alive_progress import alive_bar

        except ModuleNotFoundError:
            raise MissingDependencyException(
                "AliveCollector",
                ["alive-progress"],
                "We need alive-progress to display progress bars.",
            )
        self.step = None
        self.single_bar = single_bar
        self._bars: Dict[str, Any] = {}
        self._bars_counts: Dict[str, int] = {}
        self._bars_contexts: Dict[str, ContextManager[Any]] = {}
        self.alive_kwargs = alive_kwargs or {}

    def update(
        self,
        name: str,
        inc: int = 1,
        total: int = None,
        inc_total: int = None,
        message: str = None,
        label: str = "",
    ) -> None:
        key = f"{name}_{label}"
        bar = self._bars.get(key)
        if bar is None:
            if label:
                name = f"{name}[{label}]"
            if len(self._bars) == 0:
                desc = self.step + ": " + name
            else:
                # do not add any more counters
                if self.single_bar:
                    return
                desc = name
            bar = alive_bar(total=total, title=desc, **self.alive_kwargs)
            self._bars_contexts[key] = bar
            bar = self._bars[key] = bar.__enter__()
            self._bars_counts[key] = 0
        else:
            # TODO: implement once total change is supported
            pass

        # if message:
        #     bar.set_postfix_str(message)
        if inc > 0:
            bar(inc)
            self._bars_counts[key] += inc

    def _start(self, step: str) -> None:
        self._bars = {}
        self._bars_contexts = {}
        self

    def _stop(self) -> None:
        for bar in self._bars_contexts.values():
            bar.__exit__(None, None, None)
        self._bars.clear()
        self._bars_contexts.clear()
        self._bars_counts.clear()


class EnlightenCollector(Collector):
    """A Collector that shows progress with `enlighten` progress and status bars that also allow for logging."""

    _bars: Dict[str, EnlCounter]
    _manager: EnlManager
    _status: EnlStatusBar

    def __init__(self, single_bar: bool = False, **enlighten_kwargs: Any) -> None:
        """Collector that uses Enlighten to display counters as progress bars. Set `single_bar` to True to show just the main progress bar. Pass any config to Enlighten in kwargs"""
        try:
            global enlighten

            import enlighten
            from enlighten import (
                Counter as EnlCounter,
                StatusBar as EnlStatusBar,
                Manager as EnlManager,
            )
        except ModuleNotFoundError:
            raise MissingDependencyException(
                "EnlightenCollector",
                ["enlighten"],
                "We need enlighten to display progress bars with a space for log messages.",
            )
        self.step = None
        self.single_bar = single_bar
        self.enlighten_kwargs = enlighten_kwargs

    def update(
        self,
        name: str,
        inc: int = 1,
        total: int = None,
        inc_total: int = None,
        message: str = None,
        label: str = "",
    ) -> None:
        key = f"{name}_{label}"
        bar = self._bars.get(key)
        if bar is None:
            if label:
                name = f"{name}[{label}]"
            if len(self._bars) > 0 and self.single_bar:
                # do not add any more counters
                return
            bar = self._manager.counter(
                desc=name, total=total, leave=True, force=True, **self.enlighten_kwargs
            )
            bar.refresh()
            self._bars[key] = bar
        else:
            if inc_total:
                bar.total = bar.total + inc_total
        bar.update(inc)

    def _start(self, step: str) -> None:
        self._bars = {}
        self._manager = enlighten.get_manager(enabled=True)
        self._status = self._manager.status_bar(
            leave=True, justify=enlighten.Justify.CENTER, fill="="
        )
        self._status.update(step)

    def _stop(self) -> None:
        if self._status:
            self._status.close()
        for bar in self._bars.values():
            bar.refresh()
            bar.close()
        self._bars.clear()
        self._manager.stop()
        self._manager = None
        self._bars = None
        self._status = None


NULL_COLLECTOR = NullCollector()



from __future__ import annotations
from collections.abc import Collection

from functools import partial
from typing import overload, Union, Any, Generator, Optional, Sequence, Type, TYPE_CHECKING
from textwrap import indent
from contextlib import contextmanager
from dlt.common.utils import simple_repr, without_none

from sqlglot import maybe_parse
from sqlglot.optimizer.merge_subqueries import merge_subqueries
from sqlglot.expressions import ExpOrStr as SqlglotExprOrStr

import sqlglot.expressions as sge

import dlt
from dlt.common.destination.dataset import TFilterOperation
from dlt.common.libs.sqlglot import to_sqlglot_type, build_typed_literal, TSqlGlotDialect
from dlt.common.libs.utils import is_instance_lib
from dlt.common.schema.typing import (
    TTableSchema,
    TTableSchemaColumns,
    LOADS_TABLE_NAME,
    C_DLT_LOADS_TABLE_LOAD_ID,
    C_DLT_LOAD_ID,
    TTableReference,
    TTableReferenceStandalone,
)
from dlt.common.schema import utils as schema_utils, TSchemaTables
from dlt.common.typing import Self, TSortOrder
from dlt.common.exceptions import ValueErrorWithKnownValues
from dlt.dataset import lineage
from dlt.destinations.sql_client import SqlClientBase, WithSqlClient
from dlt.destinations.queries import bind_query, build_select_expr
from dlt.common.exceptions import MissingDependencyException
from dlt.common.destination.dataset import SupportsDataAccess


if TYPE_CHECKING:
    from dlt.common.libs.ibis import ir
    from dlt.common.libs.pandas import pandas as pd
    from dlt.common.libs.pyarrow import pyarrow as pa
    from dlt.helpers.ibis import Expr as IbisExpr


_FILTER_OP_MAP = {
    "eq": sge.EQ,
    "ne": sge.NEQ,
    "gt": sge.GT,
    "lt": sge.LT,
    "gte": sge.GTE,
    "lte": sge.LTE,
    "in": sge.In,
    "not_in": sge.Not,
}


class Relation(WithSqlClient):
    @overload
    def __init__(
        self,
        *,
        dataset: dlt.Dataset,
        query: Union[str, sge.Query],
        query_dialect: Optional[str] = None,
        _execute_raw_query: bool = False,
    ) -> None: ...

    @overload
    def __init__(
        self,
        *,
        dataset: dlt.Dataset,
        table_name: str,
    ) -> None: ...

    def __init__(
        self,
        *,
        dataset: dlt.Dataset,
        query: Optional[Union[str, sge.Query, IbisExpr]] = None,
        query_dialect: Optional[str] = None,
        table_name: Optional[str] = None,
        _execute_raw_query: bool = False,
    ) -> None:
        """Create a lazy evaluated relation for the dataset of a destination"""
        if table_name is None and query is None:
            raise ValueError(
                "`dlt.Relation` needs to receive minimally `table_name` or `query` at"
                " initialization."
            )

        self._dataset = dataset
        self._query = query
        self._query_dialect = query_dialect
        self._table_name = table_name
        self._execute_raw_query: bool = _execute_raw_query

        self._opened_sql_client: SqlClientBase[Any] = None
        self._sqlglot_expression: sge.Query = None
        self._schema: Optional[TTableSchemaColumns] = None

    def df(self, *args: Any, **kwargs: Any) -> pd.DataFrame | None:
        with self._cursor() as cursor:
            return cursor.df(*args, **kwargs)

    def arrow(self, *args: Any, **kwargs: Any) -> pa.Table | None:
        with self._cursor() as cursor:
            return cursor.arrow(*args, **kwargs)

    def fetchall(self, *args: Any, **kwargs: Any) -> list[tuple[Any, ...]]:
        with self._cursor() as cursor:
            return cursor.fetchall(*args, **kwargs)

    def fetchmany(self, *args: Any, **kwargs: Any) -> list[tuple[Any, ...]]:
        with self._cursor() as cursor:
            return cursor.fetchmany(*args, **kwargs)

    def fetchone(self, *args: Any, **kwargs: Any) -> tuple[Any, ...] | None:
        with self._cursor() as cursor:
            return cursor.fetchone(*args, **kwargs)

    def iter_df(self, *args: Any, **kwargs: Any) -> Generator[pd.DataFrame, None, None]:
        with self._cursor() as cursor:
            yield from cursor.iter_df(*args, **kwargs)

    # TODO maybe it should return record batches
    def iter_arrow(self, *args: Any, **kwargs: Any) -> Generator[pa.Table, None, None]:
        with self._cursor() as cursor:
            yield from cursor.iter_arrow(*args, **kwargs)

    def iter_fetch(self, *args: Any, **kwargs: Any) -> Generator[list[tuple[Any, ...]], None, None]:
        with self._cursor() as cursor:
            yield from cursor.iter_fetch(*args, **kwargs)

    @property
    def columns_schema(self) -> TTableSchemaColumns:
        """dlt columns schema. Convenience method for `dlt.schema["columns"]`"""
        return self.schema.get("columns", {})

    @property
    def schema(self) -> TTableSchema:
        """dlt table schema associated with the relation.

        This infers the schema from the relation's content. It's likely to include less
        information than retrieving the schema from the pipeline or the dataset if the table
        already exists.
        """
        if self._schema is None:
            schema, _ = _get_relation_output_columns_schema(
                self,
                infer_sqlglot_schema=True,
                allow_anonymous_columns=True,
                allow_partial=True,
            )
            self._schema = schema

        assert self._schema is not None
        # TODO use lineage features to propagate table-level dlt annotations
        return {"columns": self._schema}

    @schema.setter
    def schema(self, new_value: Any) -> None:
        """Disable schema setter."""
        raise NotImplementedError("Schema may not be set")

    @property
    def columns(self) -> list[str]:
        """List of column names found on the table."""
        return list(self.columns_schema.keys())

    def _ipython_key_completions_(self) -> list[str]:
        """Provide column names as completion suggestion in interactive environments."""
        return self.columns

    # TODO can we narrow type from `sge.Query` to `sge.Select`?
    @property
    def sqlglot_expression(self) -> sge.Query:
        """SQLGlot expression"""
        if isinstance(self._sqlglot_expression, sge.Query):
            return self._sqlglot_expression

        if isinstance(self._query, (str, sge.Query)):
            expression = maybe_parse(
                self._query, dialect=self._query_dialect or self.destination_dialect
            )
        elif isinstance(self._table_name, str):
            expression = build_select_expr(table_name=self._table_name)
        elif is_instance_lib(self._query, class_ref="ibis.Expr"):
            from dlt.helpers.ibis import ibis

            assert isinstance(self._query, ibis.Expr)

            from dlt.helpers.ibis import compile_ibis_to_sqlglot

            expression = compile_ibis_to_sqlglot(self._query, self.destination_dialect)
        else:
            raise RuntimeError(
                "`dlt.Relation` is missing `table_name` and `query` to resolve the SQLGlot"
                " expression. This is an unexpected error."
            )

        self._sqlglot_expression = expression
        return self._sqlglot_expression

    @property
    def sql_client(self) -> SqlClientBase[Any]:
        return self._dataset.sql_client

    @property
    def sql_client_class(self) -> Type[SqlClientBase[Any]]:
        return self._dataset.sql_client_class

    @contextmanager
    def _cursor(self) -> Generator[SupportsDataAccess, Any, Any]:
        """Gets a DBApiCursor for the current relation"""
        try:
            self._opened_sql_client = self.sql_client

            # we only compute the columns schema if we are not executing the raw query
            if self._execute_raw_query:
                columns_schema = None
            else:
                columns_schema = self.columns_schema

            # case 1: client is already opened and managed from outside
            if self.sql_client.native_connection:
                with self.sql_client.execute_query(self.to_sql()) as cursor:
                    if columns_schema:
                        cursor.columns_schema = columns_schema
                    yield cursor
            # case 2: client is not opened, we need to manage it
            else:
                with self.sql_client as client:
                    with client.execute_query(self.to_sql()) as cursor:
                        if columns_schema:
                            cursor.columns_schema = columns_schema
                        yield cursor
        finally:
            self._opened_sql_client = None

    def to_sql(self, pretty: bool = False, *, _raw_query: bool = False) -> str:
        """Get the normalize query string in the correct sql dialect for this relation"""

        if self._execute_raw_query or _raw_query:
            query = self.sqlglot_expression
        else:
            _, _qualified_query = _get_relation_output_columns_schema(self)
            query = bind_query(
                qualified_query=_qualified_query,
                sqlglot_schema=self._dataset.sqlglot_schema,
                expand_table_name=partial(
                    self.sql_client.make_qualified_table_name_path,
                    quote=False,
                    casefold=False,
                ),
                casefold_identifier=self.sql_client.capabilities.casefold_identifier,
            )

        if not isinstance(query, sge.Query):
            raise ValueError(
                f"Query `{query}` received for `{self.__class__.__name__}`. "
                "Must be an SQL SELECT statement."
            )

        return query.sql(dialect=self.destination_dialect, pretty=pretty)

    # TODO this method needs to have the same name as `dlt.extract.hints::SqlModel.query_dialect`;
    # the current implementation doesn't disambiguate "query dialect" and "destination dialect",
    # i.e., the input and output the SQL transpilation
    # These methods are called in `dlt.normalize.items_normalizers::ModelItemsNormalizer.__call__()`
    # and should be fixed; then remove this property
    @property
    def query_dialect(self) -> TSqlGlotDialect:
        return self.destination_dialect

    @property
    def destination_dialect(self) -> TSqlGlotDialect:
        """SQLGlot dialect used by the destination.

        This is the target dialect when transpiling SQL queries.
        """
        return self._dataset.destination_dialect

    def to_ibis(self) -> ir.Table:
        """Create an `ibis.Table` expression from the current relation.

        If the `dlt.Relation` was initialized with a `table_name`, it will return an
        `ibis.Table` directly. If the `dlt.Relation` was transformed via `.where()`, `.select()`,
        etc., it will apply the operations in a single step as an opaque SQLQuery Ibis operation.
        """
        from dlt.common.libs.ibis import _DltBackend

        backend = _DltBackend.from_dataset(self._dataset)

        if self._table_name:
            ibis_table = backend.table(self._table_name)
        else:
            # pass raw query before any identifiers are expanded, quoted or normalized
            ibis_table = backend.sql(self.sqlglot_expression.sql(dialect=self.destination_dialect))

        return ibis_table

    def limit(self, limit: int) -> Self:
        """Create a `Relation` using a `LIMIT` clause."""
        rel = self.__copy__()
        rel._sqlglot_expression = rel.sqlglot_expression.limit(limit)
        return rel

    def head(self, limit: int = 5) -> Self:
        """Create a `Relation` using a `LIMIT` clause. Defaults to `limit=5`

        This proxies `Relation.limit()`.
        """
        return self.limit(limit)

    def select(self, *columns: str, _allow_merge_subqueries: bool = True) -> Self:
        """Create a `Relation` with the selected columns using a `SELECT` clause."""
        proj = [sge.Column(this=sge.to_identifier(col, quoted=True)) for col in columns]
        subquery = self.sqlglot_expression.subquery()
        new_expr = sge.select(*proj).from_(subquery)
        rel = self.__copy__()
        if _allow_merge_subqueries:
            rel._sqlglot_expression = merge_subqueries(new_expr)
        else:
            rel._sqlglot_expression = new_expr
        return rel

    def order_by(self, column_name: str, direction: TSortOrder = "asc") -> Self:
        """Create a `Relation` ordering results using a `ORDER BY` clause.

        Args:
            column_name (str): The column to order by.
            direction (TSortOrder, optional): The direction to order by: "asc"/"desc". Defaults to "asc".

        Returns:
            Self: A new Relation with the `ORDER BY` clause applied.
        """
        if direction not in ["asc", "desc"]:
            raise ValueError(
                f"`{direction}` is an invalid sort order, allowed values are: `asc` and `desc`"
            )
        order_expr = sge.Ordered(
            this=sge.Column(
                this=sge.to_identifier(column_name, quoted=True),
            ),
            desc=(direction == "desc"),
        )
        rel = self.__copy__()
        rel._sqlglot_expression = rel.sqlglot_expression.order_by(order_expr)
        return rel

    # NOTE we currently force to have one column selected; we could be more flexible
    # and rewrite the query to compute the AGG of all selected columns
    # `SELECT AGG(col1), AGG(col2), ... FROM table``
    def _apply_agg(self, agg_cls: type[sge.AggFunc]) -> Self:
        """Create a `Relation` with the aggregate function applied.

        Exactly one column must be selected.
        """
        if len(self.sqlglot_expression.selects) != 1:
            raise ValueError(
                f"{agg_cls.__name__.lower()}() requires a query with exactly one select expression."
                " Consider selecting the column you want to aggregate."
            )
        selected_col = self.sqlglot_expression.selects[0]
        expr = agg_cls(this=selected_col.this if hasattr(selected_col, "this") else selected_col)
        rel = self.__copy__()
        rel.sqlglot_expression.set("expressions", [expr])
        return rel

    def max(self) -> Self:  # noqa: A003
        """Create a `Relation` with the `MAX` aggregate applied.

        Exactly one column must be selected.
        """
        return self._apply_agg(sge.Max)

    def min(self) -> Self:  # noqa: A003
        """Create a `Relation` with the `MIN` aggregate applied.

        Exactly one column must be selected.
        """
        return self._apply_agg(sge.Min)

    @overload
    def where(self, column_or_expr: SqlglotExprOrStr) -> Self: ...

    @overload
    def where(
        self,
        column_or_expr: str,
        operator: TFilterOperation,
        value: Any,
    ) -> Self: ...

    def where(
        self,
        column_or_expr: SqlglotExprOrStr,
        operator: Optional[TFilterOperation] = None,
        value: Optional[Any] = None,
    ) -> Self:
        """Create a `Relation` filtering results using a `WHERE` clause.

        This is identical to `Relation.filter()`.

        Args:
            column_name (str): The column to filter on.
            operator (TFilterOperation): The operator to use. Available operations are: eq, ne, gt, lt, gte, lte, in, not_in
            value (Any): The value to filter on.

        Returns:
            Self: A new Relation with the WHERE clause applied.
        """
        rel = self.__copy__()

        if not isinstance(rel.sqlglot_expression, sge.Select):
            raise ValueError(
                f"Query `{rel.sqlglot_expression}` received for `{rel.__class__.__name__}`. "
                "Must be an SQL SELECT statement."
            )

        if not operator and not value:
            rel._sqlglot_expression = rel.sqlglot_expression.where(
                column_or_expr, dialect=self.destination_dialect
            )
            return rel

        assert isinstance(column_or_expr, str)
        column_name = column_or_expr

        if isinstance(operator, str):
            try:
                condition_cls = _FILTER_OP_MAP[operator]
            except KeyError:
                raise ValueErrorWithKnownValues(
                    key="operator",
                    value_received=operator,
                    valid_values=list(_FILTER_OP_MAP.keys()),
                )

        sqlgot_type = to_sqlglot_type(
            dlt_type=self.columns_schema[column_name].get("data_type"),
            precision=self.columns_schema[column_name].get("precision"),
            timezone=self.columns_schema[column_name].get("timezone"),
            nullable=self.columns_schema[column_name].get("nullable"),
        )

        value_expr = build_typed_literal(value, sqlgot_type)

        column = sge.Column(this=sge.to_identifier(column_name, quoted=True))

        condition: sge.Expression = None
        if operator == "in":
            exprs = value_expr.expressions if isinstance(value_expr, sge.Tuple) else [value_expr]
            condition = sge.In(this=column, expressions=exprs)
        elif operator == "not_in":
            exprs = value_expr.expressions if isinstance(value_expr, sge.Tuple) else [value_expr]
            condition = sge.Not(this=sge.In(this=column, expressions=exprs))
        else:
            condition = condition_cls(this=column, expression=value_expr)

        rel._sqlglot_expression = rel.sqlglot_expression.where(condition)
        return rel

    @overload
    def filter(self, column_or_expr: SqlglotExprOrStr) -> Self: ...  # noqa: A003

    @overload
    def filter(  # noqa: A003
        self,
        column_or_expr: str,
        operator: TFilterOperation,
        value: Any,
    ) -> Self: ...

    def filter(  # noqa: A003
        self,
        column_or_expr: SqlglotExprOrStr,
        operator: Optional[TFilterOperation] = None,
        value: Optional[Any] = None,
    ) -> Self:
        """Create a `Relation` filtering results using a `WHERE` clause.

        This is identical to `Relation.where()`.

        Args:
            column_name (str): The column to filter on.
            operator (TFilterOperation): The operator to use. Available operations are: eq, ne, gt, lt, gte, lte, in, not_in
            value (Any): The value to filter on.

        Returns:
            Self: A new Relation with the WHERE clause applied.
        """
        if not operator and not value:
            return self.where(column_or_expr=column_or_expr)
        assert isinstance(column_or_expr, str)
        return self.where(column_or_expr=column_or_expr, operator=operator, value=value)

    # TODO could be refactored to join any column from `_dlt_loads` table
    def with_load_id_col(self) -> dlt.Relation:
        """Return the relation with the `_dlt_load_id` included.

        There are 3 cases:
        - If the relation is a root table, this is a no-op
        - If the relation has a root key, join relation to root table
        - If the relation has a parent key, iteratively join the root to the relation
        - Else raise

        This should only raise if the `dlt.Schema` was tempered, breaking the
        dlt-generated root and parent relationships.
        """
        table_schema = self._dataset.schema.tables[self._table_name]

        if self._dataset.schema.naming.normalize_identifier(C_DLT_LOAD_ID) in self.columns:
            return self
        elif schema_utils.has_column_with_prop(table_schema, "root_key"):
            return _add_load_id_via_root_key(self)
        elif schema_utils.has_column_with_prop(table_schema, "parent_key"):
            return _add_load_id_via_parent_key(self)
        else:
            raise ValueError

    def from_loads(
        self,
        load_ids: Collection[str],
        add_load_id_column: bool = False,
    ) -> dlt.Relation:
        """Filter the table to rows associated with `load_ids`.

        This resolves the `_dlt_load_id` column then filters rows of the
        current relation. `include_load_id` allows to keep the `_dlt_load_id` column
        or exclude it after filtering.
        """
        if not self._table_name:
            raise ValueError(
                "`filter_loads()` only works on relations created via .table()."
                " It can't be applied to arbitrary relation."
            )

        initial_columns = self.columns
        normalized_load_id = self._dataset.schema.naming.normalize_identifier(C_DLT_LOAD_ID)
        filtered_rel_with_load_id = self.with_load_id_col().where(
            normalized_load_id, "in", load_ids
        )
        return (
            filtered_rel_with_load_id
            if add_load_id_column
            else filtered_rel_with_load_id.select(*initial_columns, _allow_merge_subqueries=False)
        )

    # TODO move this to the WithSqlClient / data accessor mixin.
    def fetchscalar(self) -> Any:
        """Execute the relation and return the first value of first column as a Python primitive"""
        row = self.fetchmany(2)
        if not row:
            return None
        if len(row) != 1:
            raise ValueError(
                "Expected scalar result (single row, single column), got more than one row"
            )
        if len(row[0]) != 1:
            raise ValueError(
                "Expected scalar result (single row, single column), got 1 row with"
                f" {len(row[0])} columns"
            )
        return row[0][0]

    def __getitem__(self, columns: Sequence[str]) -> Self:
        """Create a new Relation with the specified columns selected.

        This proxies `Relation.select()`.
        """
        # NOTE remember that `issubclass(str, Sequence) is True`
        if isinstance(columns, str):
            columns = [columns]
        elif not isinstance(columns, Sequence):
            raise TypeError(
                f"Received value `{columns=:}` of type `{type(columns).__name__}`."
                " Valid types are: `[Sequence[str]]`"
            )

        unknown_columns = [col for col in columns if col not in self.columns]
        if unknown_columns:
            raise KeyError(
                f"Columns `{unknown_columns}` not found on dataset. Available columns:"
                f" {self.columns}"
            )

        return self.select(*columns)

    def __str__(self) -> str:
        # TODO: merge detection of "simple" transformation that preserve table schema
        msg = f"Relation query:\n{indent(self.to_sql(pretty=True), prefix='  ')}\n"
        msg += "Columns:\n"
        for column in self.columns_schema.values():
            # TODO: show x-annotation hints
            msg += f"{indent(column['name'], prefix='  ')} {column['data_type']}\n"
        return msg

    def __repr__(self) -> str:
        # schema may not be set
        kwargs = {
            "dataset": repr(self._dataset),
            "query": self.to_sql(pretty=True),
        }
        return simple_repr("dlt.Relation", **without_none(kwargs))

    def __copy__(self) -> Self:
        return self.__class__(dataset=self._dataset, query=self.sqlglot_expression)


def _get_relation_output_columns_schema(
    relation: dlt.Relation,
    *,
    infer_sqlglot_schema: bool = False,
    allow_anonymous_columns: bool = True,
    allow_partial: bool = False,
) -> tuple[TTableSchemaColumns, sge.Query]:
    columns_schema, normalized_query = lineage.compute_columns_schema(
        # use dlt schema compliant query so lineage will work correctly on non case folded identifiers
        relation.sqlglot_expression,
        relation._dataset.sqlglot_schema,
        dialect=relation.destination_dialect,
        infer_sqlglot_schema=infer_sqlglot_schema,
        allow_anonymous_columns=allow_anonymous_columns,
        allow_partial=allow_partial,
    )
    return columns_schema, normalized_query


def _add_load_id_via_root_key_query(
    table_name: str,
    root_table_name: str,
    child_root_key: str,
    root_row_key: str,
    *,
    normalized_load_id: str = C_DLT_LOAD_ID,
) -> sge.Select:
    child_table = sge.Table(
        this=sge.to_identifier(table_name, quoted=True),
        alias=sge.TableAlias(this=sge.to_identifier("child", quoted=False)),
    )
    root_table_expr = sge.Table(
        this=sge.to_identifier(root_table_name, quoted=True),
        alias=sge.TableAlias(this=sge.to_identifier("root", quoted=False)),
    )

    # Build column list: child.*, root._dlt_load_id
    columns = [
        sge.Column(table=sge.to_identifier("child", quoted=False), this=sge.Star()),
        sge.Column(
            table=sge.to_identifier("root", quoted=False),
            this=sge.to_identifier(normalized_load_id, quoted=True),
        ),
    ]

    join_condition = sge.EQ(
        this=sge.Column(
            table=sge.to_identifier("child", quoted=False),
            this=sge.to_identifier(child_root_key, quoted=True),
        ),
        expression=sge.Column(
            table=sge.to_identifier("root", quoted=False),
            this=sge.to_identifier(root_row_key, quoted=True),
        ),
    )

    query = (
        sge.Select(expressions=columns)
        .from_(child_table)
        .join(root_table_expr, on=join_condition, join_type="INNER")
    )
    return query


def _add_load_id_via_root_key(relation: dlt.Relation) -> dlt.Relation:
    """Return the input relation with the `_dlt_load_id` column added.

    This is done by joining the `root_table._dlt_id` with the `table._dlt_root_id`
    """
    origin_table_name: str = relation._table_name
    tables_schema = relation._dataset.schema.tables
    root_table = schema_utils.get_root_table(tables_schema, origin_table_name)
    root_table_name = root_table["name"]

    ref = schema_utils.create_root_child_reference(tables_schema, origin_table_name)
    child_root_key = ref["columns"][0]
    root_row_key = ref["referenced_columns"][0]

    # Construct SELECT with INNER JOIN
    query = _add_load_id_via_root_key_query(
        table_name=origin_table_name,
        root_table_name=root_table_name,
        child_root_key=child_root_key,
        root_row_key=root_row_key,
        normalized_load_id=relation._dataset.schema.naming.normalize_identifier(C_DLT_LOAD_ID),
    )

    rel = relation.__copy__()
    rel._sqlglot_expression = query
    return rel


def _add_load_id_via_parent_key_query(
    table_name: str, table_schemas: TSchemaTables, normalized_load_id: str = C_DLT_LOAD_ID
) -> sge.Select:
    # The reference_chain goes from root to the queried table
    # Each reference contains: child table -> parent table relationship
    # We need to build joins from the queried table back to root
    reference_chain = []
    root_table_name = schema_utils.get_root_table(table_schemas, table_name)["name"]
    for reference in schema_utils.get_all_parent_child_references_from_root(
        table_schemas, root_table_name
    ):
        reference_chain.append(reference)
        if reference["table"] == table_name:
            break

    queried_table = sge.Table(
        this=sge.to_identifier(table_name, quoted=True),
        alias=sge.TableAlias(this=sge.to_identifier("t0", quoted=False)),
    )

    # build final table query with all columns explicitly + _dlt_load_id from root table
    columns = [
        sge.Column(
            table=sge.to_identifier("t0", quoted=False),
            this=sge.to_identifier(col_name, quoted=True),
        )
        for col_name in table_schemas[table_name]["columns"]
    ]
    columns.append(
        sge.Column(
            table=sge.to_identifier(f"t{len(reference_chain)}", quoted=False),
            this=sge.to_identifier(normalized_load_id, quoted=True),
        )
    )
    query = sge.Select(expressions=columns).from_(queried_table)

    # loop through references from table to root to append INNER JOINs
    for i, ref in enumerate(reversed(reference_chain)):
        parent_alias = f"t{i + 1}"
        parent_table = sge.Table(
            this=sge.to_identifier(ref["referenced_table"], quoted=True),
            alias=sge.TableAlias(this=sge.to_identifier(parent_alias, quoted=False)),
        )
        join_condition = sge.EQ(
            this=sge.Column(
                table=sge.to_identifier(f"t{i}", quoted=False),
                this=sge.to_identifier(ref["columns"][0], quoted=True),
            ),
            expression=sge.Column(
                table=sge.to_identifier(parent_alias, quoted=False),
                this=sge.to_identifier(ref["referenced_columns"][0], quoted=True),
            ),
        )
        query = query.join(parent_table, on=join_condition, join_type="INNER")

    return query


def _add_load_id_via_parent_key(relation: dlt.Relation) -> dlt.Relation:
    """Return the input relation with the `_dlt_load_id` column added.

    This is done by iteratively joining the `root_table._dlt_id` with `child._dlt_parent_id`
    until the input relation is reached.
    """
    origin_table_name: str = relation._table_name
    table_schemas = relation._dataset.schema.tables

    query = _add_load_id_via_parent_key_query(
        table_name=origin_table_name,
        table_schemas=table_schemas,
        normalized_load_id=relation._dataset.schema.naming.normalize_identifier(C_DLT_LOAD_ID),
    )

    rel = relation.__copy__()
    rel._sqlglot_expression = query
    return rel



from typing import Sequence, Tuple, Optional, List, Union, cast
import operator
from dlt.common.libs.sql_alchemy import sa

from dlt.common.typing import TAnyDateTime
from dlt.common.utils import uniq_id
from dlt.common.destination import PreparedTableSchema, DestinationCapabilitiesContext
from dlt.common.schema.utils import (
    get_columns_names_with_prop,
    get_dedup_sort_tuple,
    get_first_column_name_with_prop,
    is_nested_table,
    get_validity_column_names,
    get_active_record_timestamp,
)
from dlt.common.time import ensure_pendulum_datetime_utc
from dlt.common.storages.load_package import load_package_state as current_load_package

from dlt.destinations.impl.sqlalchemy.db_api_client import SqlalchemyClient
from dlt.destinations.sql_jobs import SqlMergeFollowupJob
from dlt.destinations.utils import get_deterministic_temp_table_name


class SqlalchemyMergeFollowupJob(SqlMergeFollowupJob):
    """Uses SQLAlchemy to generate merge SQL statements.
    Result is equivalent to the SQL generated by `SqlMergeFollowupJob`
    except for delete-insert we use concrete tables instead of temporary tables.
    """

    @classmethod
    def _new_temp_table_name(cls, table_name: str, op: str, sql_client: SqlalchemyClient) -> str:  # type: ignore[override]
        # reproducible name so we know which table to drop
        return get_deterministic_temp_table_name(table_name, op)

    @classmethod
    def gen_merge_sql(
        cls,
        table_chain: Sequence[PreparedTableSchema],
        sql_client: SqlalchemyClient,  # type: ignore[override]
    ) -> List[str]:
        root_table = table_chain[0]

        root_table_obj = sql_client.get_existing_table(root_table["name"])
        staging_root_table_obj = root_table_obj.to_metadata(
            sql_client.metadata, schema=sql_client.staging_dataset_name
        )

        primary_key_names = get_columns_names_with_prop(root_table, "primary_key")
        merge_key_names = get_columns_names_with_prop(root_table, "merge_key")

        temp_metadata = sa.MetaData()

        append_fallback = (len(primary_key_names) + len(merge_key_names)) == 0

        sqla_statements = []

        if not append_fallback:
            key_clause = cls._generate_key_table_clauses(
                primary_key_names, merge_key_names, root_table_obj, staging_root_table_obj
            )

            # Generate the delete statements
            if len(table_chain) == 1 and not cls.requires_temp_table_for_delete():
                delete_statement = root_table_obj.delete().where(
                    sa.exists(
                        sa.select(sa.literal(1))
                        .where(key_clause)
                        .select_from(staging_root_table_obj)
                    )
                )
                sqla_statements.append(delete_statement)
            else:
                row_key_col_name = cls.get_row_key_col(
                    table_chain,
                    root_table,
                    sql_client.fully_qualified_dataset_name(),
                    sql_client.fully_qualified_dataset_name(staging=True),
                )
                row_key_col = root_table_obj.c[row_key_col_name]
                # Use a real table cause sqlalchemy doesn't have TEMPORARY TABLE abstractions
                delete_temp_table = sa.Table(
                    cls._new_temp_table_name(str(root_table_obj.name), "delete", sql_client),
                    temp_metadata,
                    # Give this column a fixed name to be able to reference it later
                    sa.Column("_dlt_id", row_key_col.type),
                    schema=staging_root_table_obj.schema,
                )
                # drop table if exists
                sqla_statements.append(sa.sql.ddl.DropTable(delete_temp_table, if_exists=True))
                # Add the CREATE TABLE statement
                sqla_statements.append(sa.sql.ddl.CreateTable(delete_temp_table))
                # Insert data into the "temporary" table
                insert_statement = delete_temp_table.insert().from_select(
                    [row_key_col],
                    sa.select(row_key_col).where(
                        sa.exists(
                            sa.select(sa.literal(1))
                            .where(key_clause)
                            .select_from(staging_root_table_obj)
                        )
                    ),
                )
                sqla_statements.append(insert_statement)

                for table in table_chain[1:]:
                    chain_table_obj = sql_client.get_existing_table(table["name"])
                    root_key_name = cls.get_root_key_col(
                        table_chain,
                        table,
                        sql_client.fully_qualified_dataset_name(),
                        sql_client.fully_qualified_dataset_name(staging=True),
                    )
                    root_key_col = chain_table_obj.c[root_key_name]

                    delete_statement = chain_table_obj.delete().where(
                        root_key_col.in_(sa.select(delete_temp_table.c._dlt_id))
                    )

                    sqla_statements.append(delete_statement)

                # Delete from root table
                delete_statement = root_table_obj.delete().where(
                    row_key_col.in_(sa.select(delete_temp_table.c._dlt_id))
                )
                sqla_statements.append(delete_statement)

        hard_delete_col_name, not_delete_cond = cls._get_hard_delete_col_and_cond(
            root_table,
            root_table_obj,
            invert=True,
        )

        dedup_sort = get_dedup_sort_tuple(root_table)  # column_name, 'asc' | 'desc'
        skip_dedup = root_table.get("x-stage-data-deduplicated", False)

        if len(table_chain) > 1 and (primary_key_names or hard_delete_col_name is not None):
            condition_column_names = (
                None if hard_delete_col_name is None else [hard_delete_col_name]
            )
            condition_columns = (
                [staging_root_table_obj.c[col_name] for col_name in condition_column_names]
                if condition_column_names is not None
                else []
            )

            staging_row_key_col = staging_root_table_obj.c[row_key_col_name]

            # Create the insert "temporary" table (but use a concrete table)
            insert_temp_table = sa.Table(
                cls._new_temp_table_name(str(root_table_obj.name), "insert", sql_client),
                temp_metadata,
                sa.Column(row_key_col_name, staging_row_key_col.type),
                schema=staging_root_table_obj.schema,
            )
            # drop table if exists
            sqla_statements.append(sa.sql.ddl.DropTable(insert_temp_table, if_exists=True))
            # create temp table
            create_insert_temp_table_statement = sa.sql.ddl.CreateTable(insert_temp_table)
            sqla_statements.append(create_insert_temp_table_statement)
            staging_primary_key_cols = [
                staging_root_table_obj.c[col_name] for col_name in primary_key_names
            ]

            inner_cols = [staging_row_key_col]

            if primary_key_names and not skip_dedup:
                # deduplicate if primary key present
                if dedup_sort is not None:
                    order_by_col = staging_root_table_obj.c[dedup_sort[0]]
                    order_dir_func = sa.asc if dedup_sort[1] == "asc" else sa.desc
                else:
                    order_by_col = sa.select(sa.literal(None))
                    order_dir_func = sa.asc
                if condition_columns:
                    inner_cols += condition_columns

                inner_select = sa.select(
                    sa.func.row_number()
                    .over(
                        partition_by=set(staging_primary_key_cols),
                        order_by=order_dir_func(order_by_col),
                    )
                    .label("_dlt_dedup_rn"),
                    *inner_cols,
                ).subquery()

                select_for_temp_insert = sa.select(inner_select.c[row_key_col_name]).where(
                    inner_select.c._dlt_dedup_rn == 1
                )
                hard_delete_col_name, not_delete_cond = cls._get_hard_delete_col_and_cond(
                    root_table,
                    inner_select,
                    invert=True,
                )

                if not_delete_cond is not None:
                    select_for_temp_insert = select_for_temp_insert.where(not_delete_cond)
            else:
                hard_delete_col_name, not_delete_cond = cls._get_hard_delete_col_and_cond(
                    root_table,
                    staging_root_table_obj,
                    invert=True,
                )

                select_for_temp_insert = sa.select(staging_row_key_col)
                if not_delete_cond is not None:
                    select_for_temp_insert = select_for_temp_insert.where(not_delete_cond)

            insert_into_temp_table = insert_temp_table.insert().from_select(
                [row_key_col_name], select_for_temp_insert
            )
            sqla_statements.append(insert_into_temp_table)

        # Insert from staging to dataset
        for table in table_chain:
            table_obj = sql_client.get_existing_table(table["name"])
            staging_table_obj = table_obj.to_metadata(
                sql_client.metadata, schema=sql_client.staging_dataset_name
            )
            select_sql = staging_table_obj.select()

            if (primary_key_names and len(table_chain) > 1) or (
                not primary_key_names
                and is_nested_table(table)
                and hard_delete_col_name is not None
            ):
                uniq_column_name = root_key_name if is_nested_table(table) else row_key_col_name
                uniq_column = staging_table_obj.c[uniq_column_name]
                select_sql = select_sql.where(
                    uniq_column.in_(
                        sa.select(
                            insert_temp_table.c[row_key_col_name].label(uniq_column_name)
                        ).subquery()
                    )
                )
            elif primary_key_names and len(table_chain) == 1 and not skip_dedup:
                staging_primary_key_cols = [
                    staging_table_obj.c[col_name] for col_name in primary_key_names
                ]
                if dedup_sort is not None:
                    order_by_col = staging_table_obj.c[dedup_sort[0]]
                    order_dir_func = sa.asc if dedup_sort[1] == "asc" else sa.desc
                else:
                    order_by_col = sa.select(sa.literal(None))
                    order_dir_func = sa.asc

                inner_select = sa.select(
                    staging_table_obj,
                    sa.func.row_number()
                    .over(
                        partition_by=set(staging_primary_key_cols),
                        order_by=order_dir_func(order_by_col),
                    )
                    .label("_dlt_dedup_rn"),
                ).subquery()

                select_sql = sa.select(
                    *[c for c in inner_select.c if c.name != "_dlt_dedup_rn"]
                ).where(inner_select.c._dlt_dedup_rn == 1)

                hard_delete_col_name, not_delete_cond = cls._get_hard_delete_col_and_cond(
                    root_table, inner_select, invert=True
                )

                if hard_delete_col_name is not None:
                    select_sql = select_sql.where(not_delete_cond)
            else:
                hard_delete_col_name, not_delete_cond = cls._get_hard_delete_col_and_cond(
                    root_table, staging_root_table_obj, invert=True
                )

                if hard_delete_col_name is not None:
                    select_sql = select_sql.where(not_delete_cond)

            insert_statement = table_obj.insert().from_select(
                [col.name for col in table_obj.columns], select_sql
            )
            sqla_statements.append(insert_statement)

        return [
            str(stmt.compile(sql_client.engine, compile_kwargs={"literal_binds": True}))
            for stmt in sqla_statements
        ]

    @classmethod
    def _get_hard_delete_col_and_cond(  # type: ignore[override]
        cls,
        table: PreparedTableSchema,
        table_obj: sa.Table,
        invert: bool = False,
    ) -> Tuple[Optional[str], Optional[sa.sql.elements.BinaryExpression]]:
        col_name = get_first_column_name_with_prop(table, "hard_delete")
        if col_name is None:
            return None, None
        col = table_obj.c[col_name]
        if invert:
            cond = col.is_(None)
        else:
            cond = col.isnot(None)
        if table["columns"][col_name]["data_type"] == "bool":
            # do not use is below
            if invert:
                cond = sa.or_(cond, col == False)  # noqa
            else:
                cond = col == True  # noqa
        return col_name, cond

    @classmethod
    def _generate_key_table_clauses(
        cls,
        primary_keys: Sequence[str],
        merge_keys: Sequence[str],
        root_table_obj: sa.Table,
        staging_root_table_obj: sa.Table,
    ) -> sa.sql.ClauseElement:
        # Returns an sqlalchemy or_ clause
        clauses = []
        if primary_keys or merge_keys:
            for key in primary_keys:
                clauses.append(
                    sa.and_(
                        *[
                            root_table_obj.c[key] == staging_root_table_obj.c[key]
                            for key in primary_keys
                        ]
                    )
                )
            for key in merge_keys:
                clauses.append(
                    sa.and_(
                        *[
                            root_table_obj.c[key] == staging_root_table_obj.c[key]
                            for key in merge_keys
                        ]
                    )
                )
            return sa.or_(*clauses)  # type: ignore[no-any-return]
        else:
            return sa.true()  # type: ignore[no-any-return]

    @classmethod
    def _gen_concat_sqla(
        cls, columns: Sequence[sa.Column]
    ) -> Union[sa.sql.elements.BinaryExpression, sa.Column]:
        # Use col1 + col2 + col3 ... to generate a dialect specific concat expression
        result = columns[0]
        if len(columns) == 1:
            return result
        # Cast because CONCAT is only generated for string columns
        result = sa.cast(result, sa.String)
        for col in columns[1:]:
            result = operator.add(result, sa.cast(col, sa.String))
        return result

    @classmethod
    def gen_scd2_sql(
        cls,
        table_chain: Sequence[PreparedTableSchema],
        sql_client: SqlalchemyClient,  # type: ignore[override]
    ) -> List[str]:
        sqla_statements = []
        root_table = table_chain[0]
        root_table_obj = sql_client.get_existing_table(root_table["name"])
        staging_root_table_obj = root_table_obj.to_metadata(
            sql_client.metadata, schema=sql_client.staging_dataset_name
        )

        from_, to = get_validity_column_names(root_table)
        hash_ = get_first_column_name_with_prop(root_table, "x-row-version")

        caps = sql_client.capabilities

        format_datetime_literal = caps.format_datetime_literal
        if format_datetime_literal is None:
            format_datetime_literal = (
                DestinationCapabilitiesContext.generic_capabilities().format_datetime_literal
            )
        _boundary_ts = cast(Optional[TAnyDateTime], root_table.get("x-boundary-timestamp"))
        boundary_ts: TAnyDateTime = (
            _boundary_ts
            if _boundary_ts is not None
            else current_load_package()["state"]["created_at"]
        )
        boundary_ts = ensure_pendulum_datetime_utc(boundary_ts)

        boundary_literal = format_datetime_literal(boundary_ts, caps.timestamp_precision)

        active_record_timestamp = get_active_record_timestamp(root_table)

        update_statement = (
            root_table_obj.update()
            .values({to: sa.text(boundary_literal)})
            .where(root_table_obj.c[hash_].notin_(sa.select(staging_root_table_obj.c[hash_])))
        )

        if active_record_timestamp is None:
            active_record_literal = None
            root_is_active_clause = root_table_obj.c[to].is_(None)
        else:
            active_record_literal = format_datetime_literal(
                active_record_timestamp, caps.timestamp_precision
            )
            root_is_active_clause = root_table_obj.c[to] == sa.text(active_record_literal)

        update_statement = update_statement.where(root_is_active_clause)

        merge_keys = get_columns_names_with_prop(root_table, "merge_key")
        if merge_keys:
            root_merge_key_cols = [root_table_obj.c[key] for key in merge_keys]
            staging_merge_key_cols = [staging_root_table_obj.c[key] for key in merge_keys]

            update_statement = update_statement.where(
                cls._gen_concat_sqla(root_merge_key_cols).in_(
                    sa.select(cls._gen_concat_sqla(staging_merge_key_cols))
                )
            )

        sqla_statements.append(update_statement)

        insert_statement = root_table_obj.insert().from_select(
            [col.name for col in root_table_obj.columns],
            sa.select(
                sa.literal(boundary_literal.strip("'")).label(from_),
                sa.literal(
                    active_record_literal.strip("'") if active_record_literal is not None else None
                ).label(to),
                *[c for c in staging_root_table_obj.columns if c.name not in [from_, to]],
            ).where(
                staging_root_table_obj.c[hash_].notin_(
                    sa.select(root_table_obj.c[hash_]).where(root_is_active_clause)
                )
            ),
        )
        sqla_statements.append(insert_statement)

        nested_tables = table_chain[1:]
        for table in nested_tables:
            row_key_column = cls.get_row_key_col(
                table_chain,
                table,
                sql_client.fully_qualified_dataset_name(),
                sql_client.fully_qualified_dataset_name(staging=True),
            )

            table_obj = sql_client.get_existing_table(table["name"])
            staging_table_obj = table_obj.to_metadata(
                sql_client.metadata, schema=sql_client.staging_dataset_name
            )

            insert_statement = table_obj.insert().from_select(
                [col.name for col in table_obj.columns],
                staging_table_obj.select().where(
                    staging_table_obj.c[row_key_column].notin_(
                        sa.select(table_obj.c[row_key_column])
                    )
                ),
            )
            sqla_statements.append(insert_statement)

        return [
            str(stmt.compile(sql_client.engine, compile_kwargs={"literal_binds": True}))
            for stmt in sqla_statements
        ]



from typing import Callable, Dict, Union

from pendulum.datetime import DateTime
from typing_extensions import TypeAlias


TCurrentDateTimeCallback: TypeAlias = Callable[[], DateTime]
"""A callback function to which should return pendulum.DateTime instance"""

TCurrentDateTime: TypeAlias = Union[DateTime, TCurrentDateTimeCallback]
"""pendulum.DateTime instance or a callable which should return pendulum.DateTime"""

TLayoutPlaceholderCallback: TypeAlias = Callable[[str, str, str, str, str], str]
"""A callback which should return prepared string value the following arguments passed
`schema name`, `table name`, `load_id`, `file_id` and an `extension`
"""

TExtraPlaceholders: TypeAlias = Dict[
    str, Union[Union[str, int, DateTime], TLayoutPlaceholderCallback]
]
"""Extra placeholders for filesystem layout"""



import marimo

__generated_with = "0.19.4"
app = marimo.App(width="medium", app_title="pipeline_selector")

with app.setup:
    import marimo as mo
    from dlt._workspace.cli.utils import list_local_pipelines


@app.cell
def pipeline_locations():
    _pipelines_dir, _pipelines = list_local_pipelines()
    pipelines_locations = {p["name"]: _pipelines_dir + "/" + p["name"] for p in _pipelines}
    return (pipelines_locations,)


@app.cell
def pipeline_selector(pipelines_locations):
    try:
        _first_pipeline_name = next(iter(pipelines_locations.keys()))
    except StopIteration:
        _first_pipeline_name = None

    pipeline_selector = mo.ui.dropdown(
        pipelines_locations.keys(),
        value=_first_pipeline_name,
        label="Pipeline",
    )
    pipeline_selector
    return (pipeline_selector,)


@app.cell
def outputs(pipeline_selector, pipelines_locations):
    pipeline_name = pipeline_selector.value
    pipeline_path = pipelines_locations.get(pipeline_name, None)  # noqa: F841
    return


if __name__ == "__main__":
    app.run()



import os
from importlib import import_module
from types import ModuleType
from typing import Any, Dict, Optional, Type, Tuple, cast

import dlt
from dlt.common import logger
from dlt.common import known_env
from dlt.common.configuration.inject import with_config
from dlt.common.configuration.specs import known_sections
from dlt.common.schema.configuration import SchemaConfiguration
from dlt.common.normalizers.exceptions import InvalidJsonNormalizer
from dlt.common.normalizers.json import SupportsDataItemNormalizer, DataItemNormalizer
from dlt.common.normalizers.naming import NamingConvention
from dlt.common.normalizers.naming.exceptions import (
    NamingTypeNotFound,
    UnknownNamingModule,
    InvalidNamingType,
)
from dlt.common.normalizers.typing import (
    TJSONNormalizer,
    TNormalizersConfig,
    TNamingConventionReferenceArg,
)
from dlt.common.typing import is_subclass
from dlt.common.utils import get_full_obj_class_name

DEFAULT_NAMING_NAMESPACE = os.environ.get(
    known_env.DLT_DEFAULT_NAMING_NAMESPACE, "dlt.common.normalizers.naming"
)
DEFAULT_NAMING_MODULE = os.environ.get(known_env.DLT_DEFAULT_NAMING_MODULE, "snake_case")
DEFAULT_JSON_NAMESPACE = "dlt.common.normalizers.json"
DEFAULT_JSON_MODULE = "relational"


def _section_for_schema(kwargs: Dict[str, Any]) -> Tuple[str, ...]:
    """Uses the schema name to generate dynamic section normalizer settings"""
    if schema_name := kwargs.get("schema_name"):
        return (known_sections.SOURCES, schema_name)
    else:
        return (known_sections.SOURCES,)


@with_config(spec=SchemaConfiguration, sections=_section_for_schema)  # type: ignore[call-overload]
def configured_normalizers(
    naming: TNamingConventionReferenceArg = dlt.config.value,
    json_normalizer: TJSONNormalizer = dlt.config.value,
    allow_identifier_change_on_table_with_data: bool = None,
    use_break_path_on_normalize: Optional[bool] = None,
    schema_name: Optional[str] = None,
) -> TNormalizersConfig:
    """Gets explicitly configured normalizers without any defaults or capabilities injection. If `naming`
    is a module or a type it will get converted into string form via import.

    If `schema_name` is present, a section ("sources", schema_name, "schema") is used to inject the config
    via callable sections=_section_for_schema
    """

    norm_conf: TNormalizersConfig = {"names": serialize_reference(naming), "json": json_normalizer}
    if allow_identifier_change_on_table_with_data is not None:
        norm_conf["allow_identifier_change_on_table_with_data"] = (
            allow_identifier_change_on_table_with_data
        )
    if use_break_path_on_normalize is not None:
        norm_conf["use_break_path_on_normalize"] = use_break_path_on_normalize
    return norm_conf


@with_config
def import_normalizers(
    explicit_normalizers: TNormalizersConfig,
    default_normalizers: TNormalizersConfig = None,
) -> Tuple[TNormalizersConfig, NamingConvention, Type[DataItemNormalizer[Any]]]:
    """Imports the normalizers specified in `normalizers_config` or taken from defaults. Returns the updated config and imported modules.

    `destination_capabilities` are used to get naming convention, max length of the identifier and max nesting level.
    """
    # use container to get destination capabilities, do not use config injection to resolve circular dependencies
    from dlt.common.destination.capabilities import DestinationCapabilitiesContext
    from dlt.common.configuration.container import Container

    destination_capabilities = Container().get(DestinationCapabilitiesContext)
    if default_normalizers is None:
        default_normalizers = {}
    # add defaults to normalizer_config
    naming: Optional[TNamingConventionReferenceArg] = explicit_normalizers.get("names")
    if naming is None:
        if destination_capabilities:
            naming = destination_capabilities.naming_convention
        if naming is None:
            naming = default_normalizers.get("names") or DEFAULT_NAMING_MODULE
    # get max identifier length
    if destination_capabilities:
        max_length = min(
            destination_capabilities.max_identifier_length,
            destination_capabilities.max_column_identifier_length,
        )
    else:
        max_length = None
    naming_convention = naming_from_reference(naming, max_length)
    explicit_normalizers["names"] = serialize_reference(naming)

    item_normalizer = explicit_normalizers.get("json") or default_normalizers.get("json") or {}
    item_normalizer.setdefault("module", DEFAULT_JSON_MODULE)
    # resolve shorthand module names (e.g. "relational" -> "dlt.common.normalizers.json.relational")
    json_module_name = item_normalizer["module"]
    if "." not in json_module_name:
        json_module_name = f"{DEFAULT_JSON_NAMESPACE}.{json_module_name}"
        item_normalizer["module"] = json_module_name
    # if max_table_nesting is set, we need to set the max_table_nesting in the json_normalizer
    if destination_capabilities and destination_capabilities.max_table_nesting is not None:
        # TODO: this is a hack, we need a better method to do this
        from dlt.common.normalizers.json.relational import DataItemNormalizer

        try:
            DataItemNormalizer.ensure_this_normalizer(item_normalizer)
            item_normalizer.setdefault("config", {})
            item_normalizer["config"]["max_nesting"] = destination_capabilities.max_table_nesting  # type: ignore[index]
        except InvalidJsonNormalizer:
            # not a right normalizer
            logger.warning(f"JSON Normalizer {item_normalizer} does not support max_nesting")
            pass
    json_module = cast(SupportsDataItemNormalizer, import_module(json_module_name))
    explicit_normalizers["json"] = item_normalizer
    return (
        explicit_normalizers,
        naming_convention,
        json_module.DataItemNormalizer,
    )


def naming_from_reference(
    names: TNamingConventionReferenceArg,
    max_length: Optional[int] = None,
) -> NamingConvention:
    """Resolves naming convention from reference in `names` and applies max length if specified

    Reference may be: (1) shorthand name pointing to `dlt.common.normalizers.naming` namespace
    (2) a type name which is a module containing `NamingConvention` attribute (3) a type of class deriving from NamingConvention
    """

    def _import_naming(module: str) -> ModuleType:
        if "." in module:
            # TODO: bump schema engine version and migrate schema. also change the name in  TNormalizersConfig from names to naming
            if module == "dlt.common.normalizers.names.snake_case":
                module = f"{DEFAULT_NAMING_NAMESPACE}.{DEFAULT_NAMING_MODULE}"
            # this is full module name
            naming_module = import_module(module)
        else:
            # from known location
            try:
                naming_module = import_module(f"{DEFAULT_NAMING_NAMESPACE}.{module}")
            except ImportError:
                # also import local module
                naming_module = import_module(module)
        return naming_module

    def _get_type(naming_module: ModuleType, cls: str) -> Type[NamingConvention]:
        class_: Type[NamingConvention] = getattr(naming_module, cls, None)
        if class_ is None:
            raise NamingTypeNotFound(naming_module.__name__, cls)
        if is_subclass(class_, NamingConvention):
            return class_
        raise InvalidNamingType(naming_module.__name__, cls)

    if is_subclass(names, NamingConvention):
        class_: Type[NamingConvention] = names  # type: ignore[assignment]
    elif isinstance(names, ModuleType):
        class_ = _get_type(names, "NamingConvention")
    elif isinstance(names, str):
        try:
            class_ = _get_type(_import_naming(names), "NamingConvention")
        except ImportError:
            parts = names.rsplit(".", 1)
            # we have no more options to try
            if len(parts) <= 1:
                raise UnknownNamingModule(names)
            try:
                class_ = _get_type(_import_naming(parts[0]), parts[1])
            except UnknownNamingModule:
                raise
            except ImportError:
                raise UnknownNamingModule(names)
    else:
        raise ValueError(names)

    return class_(max_length)


def serialize_reference(naming: Optional[TNamingConventionReferenceArg]) -> Optional[str]:
    """Serializes generic `naming` reference to importable string."""
    if naming is None:
        return naming
    if isinstance(naming, str):
        return naming
    # import reference and use naming to get valid path to type
    return get_full_obj_class_name(naming_from_reference(naming))



import semver
from typing import Iterable, Union, List

from dlt.common.exceptions import DltException, TerminalValueError

VersionLike = Union[semver.Version, int]


class StorageException(DltException):
    def __init__(self, msg: str) -> None:
        super().__init__(msg)


class NoMigrationPathException(StorageException):
    def __init__(
        self,
        storage_path: str,
        initial_version: VersionLike,
        migrated_version: VersionLike,
        target_version: VersionLike,
    ) -> None:
        self.storage_path = storage_path
        self.initial_version = initial_version
        self.migrated_version = migrated_version
        self.target_version = target_version
        super().__init__(
            f"Could not find migration path for `{storage_path}` from `{initial_version=:}` to"
            f" `{target_version=:}`, stopped at `{migrated_version}`"
        )


class WrongStorageVersionException(StorageException):
    def __init__(
        self,
        storage_path: str,
        initial_version: semver.Version,
        target_version: semver.Version,
    ) -> None:
        self.storage_path = storage_path
        self.initial_version = initial_version
        self.target_version = target_version
        super().__init__(
            f"Expected storage `{storage_path}` with `{target_version=:}` but found"
            f" `{initial_version=:}`"
        )


class UnsupportedStorageVersionException(StorageException):
    def __init__(
        self,
        storage_path: str,
        initial_version: int,
        current_version: int,
        supported_versions: set[int],
    ) -> None:
        self.storage_path = storage_path
        self.initial_version = initial_version
        self.current_version = current_version
        self.supported_versions = supported_versions
        supported_str = ", ".join(str(v) for v in supported_versions)
        super().__init__(
            f"Expected storage `{storage_path}` with version in [{supported_str}] "
            f"but found `{initial_version=}`; `{current_version=}`."
        )


class StorageMigrationError(StorageException):
    def __init__(
        self,
        storage_path: str,
        from_version: semver.Version,
        target_version: semver.Version,
        info: str,
    ) -> None:
        self.storage_path = storage_path
        self.from_version = from_version
        self.target_version = target_version
        super().__init__(
            f"Storage `{storage_path}` with `{target_version=:}` at `{from_version}`: " + info
        )


class LoadStorageException(StorageException):
    pass


class JobFileFormatUnsupported(LoadStorageException, TerminalValueError):
    def __init__(self, load_id: str, supported_formats: Iterable[str], wrong_job: str) -> None:
        self.load_id = load_id
        self.expected_file_formats = supported_formats
        self.wrong_job = wrong_job
        super().__init__(
            f"Job `{wrong_job}` for load id `{load_id}` requires job file format that is not one"
            f" of: {supported_formats}"
        )


class LoadPackageNotFound(LoadStorageException, FileNotFoundError):
    def __init__(self, load_id: str) -> None:
        self.load_id = load_id
        super().__init__(f"Package with `{load_id=:}` could not be found")


class LoadPackageCancelled(LoadStorageException):
    def __init__(self, load_id: str) -> None:
        self.load_id = load_id
        super().__init__(f"Package with `{load_id=:}` is cancelled")


class LoadPackageAlreadyCompleted(LoadStorageException):
    def __init__(self, load_id: str) -> None:
        self.load_id = load_id
        super().__init__(
            f"Package with `{load_id=:}` is already completed, but another complete was requested"
        )


class LoadPackageNotCompleted(LoadStorageException):
    def __init__(self, load_id: str) -> None:
        self.load_id = load_id
        super().__init__(
            f"Package with `{load_id=:}` is not yet completed, but the calling function required a"
            " complete package."
        )


class SchemaStorageException(StorageException):
    pass


class InStorageSchemaModified(SchemaStorageException):
    def __init__(self, schema_name: str, storage_path: str) -> None:
        msg = (
            f"Schema `{schema_name}` in `{storage_path}` was externally modified. This is not"
            " allowed as that would prevent correct version tracking. Use import/export"
            " capabilities of dlt to provide external changes."
        )
        super().__init__(msg)


class SchemaNotFoundError(SchemaStorageException, FileNotFoundError, KeyError):
    def __init__(
        self,
        schema_name: str,
        storage_path: str,
        import_path: str = None,
        import_format: str = None,
    ) -> None:
        msg = f"Schema `{schema_name}` in `{storage_path}` could not be found."
        if import_path:
            msg += f"Failed to import using `{import_path=:}` and `{import_format=:}`."
        super().__init__(msg)


class UnexpectedSchemaName(SchemaStorageException, ValueError):
    def __init__(self, schema_name: str, storage_path: str, stored_name: str) -> None:
        super().__init__(
            f"A schema file name `{schema_name}` in `{storage_path}` does not correspond to the"
            f" name of schema in the file `{stored_name}`"
        )


class CurrentLoadPackageStateNotAvailable(StorageException):
    def __init__(self) -> None:
        super().__init__(
            "State of the current load package is not available. Current load package state is"
            " only available in a function decorated with `@dlt.destination` during loading."
        )



from contextlib import contextmanager, suppress
from typing import Any, AnyStr, ClassVar, Generator, Iterator, Optional, Sequence

import snowflake.connector as snowflake_lib

from dlt.common.destination import DestinationCapabilitiesContext
from dlt.destinations.exceptions import (
    DatabaseTerminalException,
    DatabaseTransientException,
    DatabaseUndefinedRelation,
)
from dlt.destinations.sql_client import (
    DBApiCursorImpl,
    SqlClientBase,
    TJobQueryTags,
    raise_database_error,
    raise_open_connection_error,
)
from dlt.destinations.typing import DBApi, DBTransaction, DataFrame, ArrowTable
from dlt.destinations.impl.snowflake.configuration import SnowflakeCredentials
from dlt.common.destination.dataset import DBApiCursor


class SnowflakeCursorImpl(DBApiCursorImpl):
    native_cursor: snowflake_lib.cursor.SnowflakeCursor  # type: ignore[assignment]

    def iter_df(self, chunk_size: int) -> Generator[DataFrame, None, None]:
        # full frame
        if not chunk_size:
            yield self.native_cursor.fetch_pandas_all()
            return
        # iterate
        # NOTE: no way to impact chunk size
        yield from self.native_cursor.fetch_pandas_batches()

    def iter_arrow(self, chunk_size: int) -> Generator[ArrowTable, None, None]:
        # TODO: figure out if empty table should be returned (there's a test for that)
        if not chunk_size:
            yield self.native_cursor.fetch_arrow_all(force_return_table=False)
            return
        # iterate
        # NOTE: no way to impact chunk size
        yield from self.native_cursor.fetch_arrow_batches()


class SnowflakeSqlClient(SqlClientBase[snowflake_lib.SnowflakeConnection], DBTransaction):
    dbapi: ClassVar[DBApi] = snowflake_lib

    def __init__(
        self,
        dataset_name: str,
        staging_dataset_name: str,
        credentials: SnowflakeCredentials,
        capabilities: DestinationCapabilitiesContext,
        query_tag: Optional[str] = None,
    ) -> None:
        super().__init__(credentials.database, dataset_name, staging_dataset_name, capabilities)
        self._conn: snowflake_lib.SnowflakeConnection = None
        self.credentials = credentials
        self.query_tag = query_tag

    def open_connection(self) -> snowflake_lib.SnowflakeConnection:
        conn_params = self.credentials.to_connector_params()
        # set the timezone to UTC so when loading from file formats that do not have timezones
        # we get dlt expected UTC
        if "timezone" not in conn_params:
            conn_params["timezone"] = "UTC"
        if "arrow_number_to_decimal" not in conn_params:
            conn_params["arrow_number_to_decimal"] = True
        # set autocommit when opening connection to override account and user level setting
        conn_params["autocommit"] = True
        self._conn = snowflake_lib.connect(
            schema=self.fully_qualified_dataset_name(), **conn_params
        )
        return self._conn

    @raise_open_connection_error
    def close_connection(self) -> None:
        if self._conn:
            self._conn.close()
            self._conn = None

    @contextmanager
    def begin_transaction(self) -> Iterator[DBTransaction]:
        try:
            self._conn.autocommit(False)
            yield self
            self.commit_transaction()
        except Exception:
            self.rollback_transaction()
            raise

    @raise_database_error
    def commit_transaction(self) -> None:
        self._conn.commit()
        self._conn.autocommit(True)

    @raise_database_error
    def rollback_transaction(self) -> None:
        self._conn.rollback()
        self._conn.autocommit(True)

    @property
    def native_connection(self) -> "snowflake_lib.SnowflakeConnection":
        return self._conn

    def drop_tables(self, *tables: str) -> None:
        # Tables are drop with `IF EXISTS`, but snowflake raises when the schema doesn't exist.
        # Multi statement exec is safe and the error can be ignored since all tables are in the same schema.
        with suppress(DatabaseUndefinedRelation):
            super().drop_tables(*tables)

    def execute_sql(
        self, sql: AnyStr, *args: Any, **kwargs: Any
    ) -> Optional[Sequence[Sequence[Any]]]:
        with self.execute_query(sql, *args, **kwargs) as curr:
            if curr.description is None:
                return None
            else:
                f = curr.fetchall()
                return f

    @contextmanager
    @raise_database_error
    def execute_query(self, query: AnyStr, *args: Any, **kwargs: Any) -> Iterator[DBApiCursor]:
        curr: DBApiCursor = None
        db_args = args if args else kwargs if kwargs else None
        with self._conn.cursor() as curr:  # type: ignore[assignment]
            try:
                curr.execute(query, db_args, num_statements=0)
                yield SnowflakeCursorImpl(curr)
            except snowflake_lib.Error:
                try:
                    self._reset_connection()
                except snowflake_lib.Error:
                    self.close_connection()
                    self.open_connection()
                raise

    def _reset_connection(self) -> None:
        self._conn.rollback()
        self._conn.autocommit(True)

    def set_query_tags(self, tags: TJobQueryTags) -> None:
        super().set_query_tags(tags)
        if self.query_tag:
            self._tag_session()

    def _tag_session(self) -> None:
        """Wraps query with Snowflake query tag"""
        if self._query_tags:
            tag = self.query_tag.format(**self._query_tags)
            tag_query = f"ALTER SESSION SET QUERY_TAG = '{tag}'"
        else:
            tag_query = "ALTER SESSION UNSET QUERY_TAG"
        self.execute_sql(tag_query)

    @classmethod
    def _make_database_exception(cls, ex: Exception) -> Exception:
        if isinstance(ex, snowflake_lib.errors.ProgrammingError):
            if ex.sqlstate == "P0000" and ex.errno == 100132:
                # Error in a multi statement execution. These don't show the original error codes
                msg = str(ex)
                if "NULL result in a non-nullable column" in msg:
                    return DatabaseTerminalException(ex)
                elif "does not exist or not authorized" in msg:  # E.g. schema not found
                    return DatabaseUndefinedRelation(ex)
                else:
                    return DatabaseTransientException(ex)
            if ex.sqlstate in {"42S02", "02000"}:
                return DatabaseUndefinedRelation(ex)
            elif ex.sqlstate == "22023":  # Adding non-nullable no-default column
                return DatabaseTerminalException(ex)
            elif ex.sqlstate == "42000" and ex.errno == 904:  # Invalid identifier
                return DatabaseTerminalException(ex)
            elif ex.sqlstate == "22000":
                return DatabaseTerminalException(ex)
            else:
                return DatabaseTransientException(ex)

        elif isinstance(ex, snowflake_lib.errors.IntegrityError):
            return DatabaseTerminalException(ex)
        elif isinstance(ex, snowflake_lib.errors.DatabaseError):
            return DatabaseTransientException(ex)
        elif isinstance(ex, TypeError):
            # snowflake raises TypeError on malformed query parameters
            return DatabaseTransientException(snowflake_lib.errors.ProgrammingError(str(ex)))
        elif cls.is_dbapi_exception(ex):
            return DatabaseTransientException(ex)
        else:
            return ex

    @staticmethod
    def is_dbapi_exception(ex: Exception) -> bool:
        return isinstance(ex, snowflake_lib.DatabaseError)



from typing import Sequence
from dlt.common.destination.exceptions import (
    DestinationTerminalException,
    DestinationTransientException,
    WithJobError,
)


class LoadClientJobException(Exception, WithJobError):
    client_exception: BaseException


class LoadClientJobFailed(DestinationTerminalException, LoadClientJobException):
    def __init__(
        self, load_id: str, job_id: str, failed_message: str, exception: BaseException
    ) -> None:
        self.load_id = load_id
        self.job_id = job_id
        self.failed_message = failed_message
        self.client_exception = exception
        super().__init__(
            f"Job with `{job_id=:}` and `{load_id=:}` failed terminally with message:"
            f" {failed_message}. The package is aborted and cannot be retried."
        )


class LoadClientJobRetry(DestinationTransientException, LoadClientJobException):
    def __init__(
        self,
        load_id: str,
        job_id: str,
        retry_count: int,
        max_retry_count: int,
        failed_message: str,
        exception: BaseException,
    ) -> None:
        self.load_id = load_id
        self.job_id = job_id
        self.retry_count = retry_count
        self.max_retry_count = max_retry_count
        self.failed_message = failed_message
        self.client_exception = exception
        super().__init__(
            f"Job with `{job_id=:}` had {retry_count} retries which is a multiple of"
            f" `{max_retry_count=:}`. Exiting retry loop. You can still rerun the load package to"
            f" retry this job. Last failure message was: {failed_message}"
        )


class LoadClientUnsupportedFileFormats(DestinationTerminalException):
    def __init__(
        self, file_format: str, supported_file_format: Sequence[str], file_path: str
    ) -> None:
        self.file_format = file_format
        self.supported_types = supported_file_format
        self.file_path = file_path
        super().__init__(
            f"Loader does not support writer for `{file_format=:}` in  file `{file_path}`."
            f" Supported writers: `{supported_file_format}`"
        )


class LoadClientUnsupportedWriteDisposition(DestinationTerminalException):
    def __init__(self, table_name: str, write_disposition: str, file_name: str) -> None:
        self.table_name = table_name
        self.write_disposition = write_disposition
        self.file_name = file_name
        super().__init__(
            f"Loader does not support `{write_disposition=:}` in table `{table_name}` when loading"
            f" file `{file_name}`"
        )


class FollowupJobCreationFailedException(DestinationTransientException):
    def __init__(self, job_id: str) -> None:
        self.job_id = job_id
        super().__init__(f"Failed to create followup job for job with `{job_id=:}`")


class TableChainFollowupJobCreationFailedException(DestinationTransientException):
    def __init__(self, root_table_name: str, details: str) -> None:
        self.root_table_name = root_table_name
        super().__init__(
            "Failed creating table chain followup jobs for table chain with root table"
            f" `{root_table_name}`:\n{details}"
        )
